﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 10.11.2018
 * Time: 20:08
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Reflection;
using RimWorld;
using RimWorld.Planet;
using UnityEngine;
using Verse;

namespace RIMMSqol
{
	/// <summary>
	/// Description of Dialog_SelectWorldTile.
	/// </summary>
	[StaticConstructorOnStartup]
	public class Dialog_SelectWorldTile : Window
	{
		public static Predicate<int> tileFilterNoImpassable = x => !Find.World.Impassable(x);
		protected static WITab_Terrain tabTerrain = (WITab_Terrain)((WITab[])typeof(WorldInspectPane).GetField("TileTabs", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null))[0];
		protected static Material hexagonCutout = null;
		protected const float tabTerrainWidth = 440f;
		
		protected int selectedTile, currentTile, maxDistance;
		protected List<KeyValuePair<int,Vector2>> neighborTiles;
		protected Func<int,Tile,bool> onSelect;
		protected Predicate<int> tileFilter;
		protected Rect lastViewect;
		
		public override Vector2 InitialSize {
			get {
				return new Vector2(540f+tabTerrainWidth,540f);
			}
		}
		
		public Dialog_SelectWorldTile(int currentTile, int maxDistance = 7, string title = null, Func<int,Tile,bool> onSelect = null, Predicate<int> tileFilter = null) {
			base.optionalTitle = title;
			this.doCloseX = true;
			this.onlyOneOfTypeAllowed = true;
			this.absorbInputAroundWindow = true;
			this.currentTile = currentTile;
			this.maxDistance = maxDistance;
			this.onSelect = onSelect;
			if ( tileFilter == null ) this.tileFilter = tileFilterNoImpassable;
			else this.tileFilter = tileFilter;
		}
		
		override public void PreOpen() {
			base.PreOpen();
			this.selectedTile = -1;
			resetNeighbors();
		}
		
		public override void PostClose() {
			base.PostClose();
			UI.UnfocusCurrentControl();
			neighborTiles.Clear();
		}
		
		public Predicate<int> TileFilter {
			set { 
				if ( this.tileFilter != value && (value != null || this.tileFilter != tileFilterNoImpassable) ) {
					if ( value == null ) this.tileFilter = tileFilterNoImpassable;
					else this.tileFilter = value;
					this.selectedTile = -1;
					if ( IsOpen ) resetNeighbors();
				}
			}
			get { return this.tileFilter; }
		}
		
		public int CurrentTile {
			set { 
				if ( this.currentTile != value ) {
					this.currentTile = value;
					this.selectedTile = -1;
					if ( IsOpen ) resetNeighbors();
				}
			}
			get { return this.currentTile; }
		}
		
		protected void resetNeighbors() {
			this.neighborTiles = new List<KeyValuePair<int,Vector2>>();
			Vector2 drawPosCurrentOnScreen = GenWorldUI.WorldToUIPosition(Find.WorldGrid.GetTileCenter(currentTile));
			Vector2 vScale = Vector2.one;
			float scale = 0f;
			Find.WorldFloodFiller.FloodFill(currentTile, tileFilter, delegate(int tile, int dist) {
            	if (dist > maxDistance) {
					return true;
				}
            	Vector2 vOff = (drawPosCurrentOnScreen-GenWorldUI.WorldToUIPosition(Find.WorldGrid.GetTileCenter(tile)));
            	if ( neighborTiles.Count == 1 ) {
            		scale = -32f / Mathf.Sqrt(vOff.sqrMagnitude); //normalizing the current zoom of the world camera
            		vScale = new Vector2(scale,scale);
            	}
            	vOff.Scale(vScale);
            	neighborTiles.Add(new KeyValuePair<int, Vector2>(tile,vOff));
				return false;
			});
		}
		
		public override void DoWindowContents(Rect inRect) {
			Vector2 sizeHex = new Vector2(32f,32f);
			Vector2 sizeHexInner = new Vector2(10f/16f*sizeHex.x,10f/16f*sizeHex.y);
			Vector2 center = inRect.center; center.x += tabTerrainWidth / 2f;
			WorldGrid grid = Find.World.grid;
			float tooltipDistance = (sizeHex.sqrMagnitude / 4f) * 0.6f;
			foreach ( KeyValuePair<int,Vector2> p in neighborTiles ) {
				int neighborTile = p.Key;
				Tile t = grid[neighborTile];
				Vector2 offsetCenter = center+p.Value;
				
				if ( tooltipDistance > (offsetCenter-Event.current.mousePosition).sqrMagnitude && !Mouse.IsInputBlockedNow ) {
					int oldTile = Find.WorldSelector.selectedTile;
					Find.WorldSelector.selectedTile = neighborTile;
					typeof(WITab_Terrain).GetMethod("FillTab", BindingFlags.Instance | BindingFlags.NonPublic).Invoke(tabTerrain,new object[0]);
					Find.WorldSelector.selectedTile = oldTile;
				}
				
				if (Event.current.type == EventType.Repaint) {
					//delaying the initialization ensures we are in the right thread
					if ( hexagonCutout == null ) {
						MaterialRequest req2 = default(MaterialRequest);
						req2.mainTex = Materials.hexagonRotationWorld;
						req2.shader = ShaderTypeDefOf.CutoutComplex.Shader;
						req2.color = Color.clear;
						req2.colorTwo = Color.white;
						req2.renderQueue = 3500;
						req2.shaderParameters = null;
						req2.maskTex = Materials.hexagonRotationWorldMask;
						hexagonCutout = MaterialPool.MatFrom(req2);
					}
					
					GenUI.DrawTextureWithMaterial(new Rect(offsetCenter-sizeHex/2f,sizeHex),(Texture2D)t.biome.DrawMaterial.mainTexture,hexagonCutout);
					Rect rectTileMarker = new Rect(offsetCenter-sizeHexInner/2f,sizeHexInner);
					foreach ( WorldObject wo in Find.WorldObjects.ObjectsAt(neighborTile) ) {
						GenUI.DrawTextureWithMaterial(rectTileMarker,wo.Material.mainTexture,wo.Material);
					}
				} else if (Event.current.rawType == EventType.MouseUp) {
					if ( tooltipDistance > (offsetCenter-Event.current.mousePosition).sqrMagnitude && !Mouse.IsInputBlockedNow ) {
						if (Event.current.button == 0) {
							Event.current.Use();
							if ( onSelect == null || onSelect(neighborTile,t) ) {
								this.selectedTile = neighborTile;
								Find.WindowStack.TryRemove(this, true);
								break;
							}
						}
					}
				}
			}
		}
	}
}
