﻿using HarmonyLib;
using RIMMSqol.translations;
using RimWorld;
using RimWorld.Planet;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UnityEngine;
using Verse;

namespace RIMMSqol
{
    [HarmonyPatch(typeof(GlobalControlsUtility))]
    [HarmonyPatch("DoDate")]
    [HarmonyPatchNamespace("GlobalControls")]
    static class GlobalControlsUtility_DoDate
    {
        private static bool useShortform = true, useMinutes = true;
        private static readonly List<string> fastHourStrings = new List<string>();
        private static readonly List<string> fastMinuteStrings = new List<string>();
        private static readonly List<string> seasonsCached = new List<string>();
        private static readonly float minuteConversion = 60f / 2500f;

        static void Prepare()
        {
            QOLMod.addApplySettingsListener(mod =>
            {
                ReadSettings();
                RebuildCache();
            });
            ReadSettings();

            LanguageChangeListener.translationActions.Add(() => RebuildCache());
        }

        static void ReadSettings()
        {
            DateStyleDef style = QOLMod.getSettings().dateStyle;
            if (style != null)
            {
                useShortform = style.useShortform;
                useMinutes = style.useMinutes;
            }
            else
            {
                useShortform = false;
                useMinutes = false;
            }
        }

        static void RebuildCache()
        {
            fastHourStrings.Clear();
            fastMinuteStrings.Clear();
            seasonsCached.Clear();
            if (useShortform)
            {
                if (!useMinutes)
                {
                    string hourLetter = "LetterHour".Translate();
                    for (int i = 0; i < 24; i++)
                    {
                        fastHourStrings.Add(new StringBuilder().Append(i).Append(hourLetter).Append(", ").ToString());
                    }
                }
                else
                {
                    for (int i = 0; i < 24; i++)
                    {
                        fastHourStrings.Add(i.ToString("00"));
                    }
                }
                if (useMinutes)
                {
                    for (int i = 0; i < 60; i++)
                    {
                        fastMinuteStrings.Add(i.ToString(":00\\, "));
                    }
                }
                int length = Enum.GetValues(typeof(Season)).Length;
                for (int j = 0; j < length; j++)
                {
                    Season season = (Season)j;
                    seasonsCached.Add((season == Season.Undefined) ? "" : new StringBuilder().Append(" ").Append(season.LabelCap()).ToString());
                }
            }
        }

        [HarmonyPriority(Priority.Last)]
        static bool Prefix(float leftX, float width, ref float curBaseY)
        {
            if (useShortform)
            {
                Vector2 vector;
                if (WorldRendererUtility.WorldRenderedNow && Find.WorldSelector.selectedTile >= 0)
                {
                    vector = Find.WorldGrid.LongLatOf(Find.WorldSelector.selectedTile);
                }
                else if (WorldRendererUtility.WorldRenderedNow && Find.WorldSelector.NumSelectedObjects > 0)
                {
                    vector = Find.WorldGrid.LongLatOf(Find.WorldSelector.FirstSelectedObject.Tile);
                }
                else
                {
                    if (Find.CurrentMap == null)
                    {
                        return false;
                    }
                    vector = Find.WorldGrid.LongLatOf(Find.CurrentMap.Tile);
                }
                int index = GenDate.HourInteger((long)Find.TickManager.TicksAbs, vector.x);
                int num = GenDate.DayOfTwelfth((long)Find.TickManager.TicksAbs, vector.x);
                Season season = GenDate.Season((long)Find.TickManager.TicksAbs, vector);
                Quadrum quadrum = GenDate.Quadrum((long)Find.TickManager.TicksAbs, vector.x);
                int num2 = GenDate.Year((long)Find.TickManager.TicksAbs, vector.x);
                string seasonLabel = IsSeasonLabelVisible ? seasonsCached[(int)season] : "";
                StringBuilder sb = new StringBuilder();

                string time;
                if (useMinutes)
                {
                    int minutes = Mathf.FloorToInt(((Find.TickManager.TicksAbs + GenDate.LocalTicksOffsetFromLongitude(vector.x)) * minuteConversion) % 60);
                    time = sb.Append(fastHourStrings[index]).Append(fastMinuteStrings[minutes]).ToString();
                    sb.Clear();
                }
                else time = fastHourStrings[index];


                //Format: 6h, 12.2.5500 Spring
                //Format + Minutes: 06:07, 12.2.5500 Spring
                string fullDateString = sb
                    .Append(time)
                    .Append(DateReadoutStringAt((long)Find.TickManager.TicksAbs, vector))
                    .Append(seasonLabel).ToString();

                Text.Font = GameFont.Small;
                Text.Anchor = TextAnchor.MiddleRight;
                Vector2 labelSize = Text.CalcSize(fullDateString);

                Rect rect = new Rect(leftX + width - labelSize.x, curBaseY - labelSize.y, labelSize.x, labelSize.y);
                if (Mouse.IsOver(rect))
                {
                    Widgets.DrawHighlight(rect);
                }
                Widgets.Label(rect, fullDateString);

                if (Mouse.IsOver(rect))
                {
                    StringBuilder stringBuilder = new StringBuilder();
                    for (int i = 0; i < 4; i++)
                    {
                        Quadrum quadrum2 = (Quadrum)i;
                        stringBuilder.AppendLine(quadrum2.Label() + " - " + quadrum2.GetSeason(vector.y).LabelCap());
                    }
                    TaggedString taggedString = "DateReadoutTip".Translate(GenDate.DaysPassed, 15, season.LabelCap(), 15, GenDate.Quadrum((long)GenTicks.TicksAbs, vector.x).Label(), stringBuilder.ToString());
                    TooltipHandler.TipRegion(rect, new TipSignal(taggedString, 86423));
                }

                curBaseY -= rect.height;
                return false;
            }
            return true;
        }

        private static string DateReadoutStringAt(long absTicks, Vector2 location)
        {
            return "DateReadoutShort".Translate(
                GenDate.DayOfSeason(absTicks, location.x) + 1, 
                (byte)GenDate.Quadrum(absTicks, location.x) + 1, 
                GenDate.Year(absTicks, location.x));
        }

        private static bool IsSeasonLabelVisible
        {
            get => !WorldRendererUtility.WorldRenderedNow && Find.CurrentMap != null;
        }
    }
}
