﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 30.10.2017
 * Time: 06:19
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections;
using System.Diagnostics;
using System.Linq;
using System.Collections.Generic;
using System.Reflection;
using System.Text.RegularExpressions;
using RIMMSqol.renderers;
using UnityEngine;
using Verse;

namespace RIMMSqol.genericSettings
{
	public interface ISettingsFieldProperties {
		string id{get;}
		string label{get;set;}
		string labelTooltip{get;set;}
        Func<Flow, string> labelTooltipProducer { get; set; }
        int order{get;}
		Func<object,object> getFromBaseObject{get;}
		object defaultValue{get;}
		Type type{get;}
		bool isSettings{get;}
		bool isList{get;}
		bool isSelectable{get;}
		Func<Flow,bool> isHidden{get;}
		object deepCopyValue(object obj);
	}
	
	public interface ISettingsFieldPropertiesSelectSingle : ISettingsFieldProperties {
		Func<Flow,Func<object,bool>> onSelectItem{get;}
		Func<Flow,IEnumerable<object>> selectableItems{get;}
		Func<object, String> labelProducer {get;}
        Func<object, Texture2D> iconProducer { get; }
        bool isNullAllowed{get;}
	}
	
	public interface ISettingsFieldPropertiesList : ISettingsFieldProperties {
		Func<object> createNewEntry{get;}
		bool isRemovalAllowed{get;}
		bool isAddingAllowed{get;}
		bool isListSettings{get;}
		bool isListPrimitive{get;}
	}
	
	public abstract class SettingsFieldProperties : ISettingsFieldProperties {
		virtual public bool isSettings { get { throw new NotImplementedException(); } }
		virtual public bool isList { get { throw new NotImplementedException(); } }
		virtual public bool isSelectable { get { throw new NotImplementedException(); } }
		virtual public object deepCopyValue(object obj) { return obj; }
		public string id { get; protected set; }
		public string label { get; set; }
		public string labelTooltip { get; set; }
        public Func<Flow, string> labelTooltipProducer { get; set; }
        public int order { get; protected set; }
		public Func<object, object> getFromBaseObject { get; protected set; }
		public object defaultValue { get; protected set; }
		public Type type { get; protected set; }
		public Func<Flow,bool> isHidden{get; protected set;}
		
		protected SettingsFieldProperties(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject)
		{
			if (id == null)
				throw new ArgumentNullException("id");
			if (order < 0 )
				throw new ArgumentOutOfRangeException("order", order, "Value must be greater than 0");
			if (type == null)
				throw new ArgumentNullException("type");
			if (getFromBaseObject == null)
				throw new ArgumentNullException("getFromBaseObject");
			if (defaultValue != null && !type.IsInstanceOfType(defaultValue) )
				throw new ArgumentException("defaultValue \""+defaultValue+"\" of type \""+defaultValue.GetType()+"\" is incompatible with provided field type \""+type+"\".");
			this.id = id;
			this.label = label;
			this.order = order;
			this.type = type;
			this.defaultValue = defaultValue;
			this.getFromBaseObject = getFromBaseObject;
			this.isHidden = null;
		}
		
		public SettingsFieldProperties setHidden(Func<Flow,bool> isHidden) {
			this.isHidden = isHidden;
			return this;
		}
	}
	
	public abstract class SettingsFieldPropertiesPrimitive : SettingsFieldProperties {
		override public bool isSettings { get { return false; } }
		override public bool isList { get { return false; } }
		override public bool isSelectable { get { return false; } }
		
		protected SettingsFieldPropertiesPrimitive(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject) 
			: base(id,label,order,type,defaultValue,getFromBaseObject)
		{
			if (!(type == typeof(float) || type == typeof(int) || type == typeof(string) || type == typeof(bool) || type == typeof(Color)))
				throw new ArgumentException("type for a primitive field must be either float, int, string, bool or Color.");
		}
	}
	
	public interface ISettingsFieldPropertiesPrimitiveNumeric<T> {
		T maxValue {get;}
		T minValue {get;}
	}
	
	public interface ISettingsFieldPropertiesPrimitiveInt : ISettingsFieldPropertiesPrimitiveNumeric<int> {}
	
	public class SettingsFieldPropertiesPrimitiveInt<T> : SettingsFieldPropertiesPrimitive, ISettingsFieldPropertiesPrimitiveInt {
		public int maxValue { get; protected set; }
		public int minValue { get; protected set; }
		public SettingsFieldPropertiesPrimitiveInt(string id, string label, int order, int defaultValue, Func<T, int> getFromBaseObject, int minValue = int.MinValue, int maxValue = int.MaxValue) 
			: base(id,label,order,typeof(int),defaultValue,o=>getFromBaseObject((T)o)) {
			this.minValue = minValue;
			this.maxValue = maxValue;
		}
	}
	
	public interface ISettingsFieldPropertiesPrimitiveFloat : ISettingsFieldPropertiesPrimitiveNumeric<float> {}
	
	public class SettingsFieldPropertiesPrimitiveFloat<T> : SettingsFieldPropertiesPrimitive, ISettingsFieldPropertiesPrimitiveFloat {
		public float maxValue { get; protected set; }
		public float minValue { get; protected set; }
		public SettingsFieldPropertiesPrimitiveFloat(string id, string label, int order, float defaultValue, Func<T, float> getFromBaseObject, float minValue = float.MinValue, float maxValue = float.MaxValue) 
			: base(id,label,order,typeof(float),defaultValue,o=>getFromBaseObject((T)o)) {
			this.minValue = minValue;
			this.maxValue = maxValue;
		}
	}
	
	public interface ISettingsFieldPropertiesPrimitiveString {
		int maxNumOfChars {get;}
		Regex inputValidator {get;}
        bool EmptyIsNull { get; }
    }
	
	public class SettingsFieldPropertiesPrimitiveString<T> : SettingsFieldPropertiesPrimitive, ISettingsFieldPropertiesPrimitiveString {
		public int maxNumOfChars { get; protected set; }
		public Regex inputValidator { get; protected set; }
        public bool EmptyIsNull { get; set; } = false;
        public SettingsFieldPropertiesPrimitiveString(string id, string label, int order, string defaultValue, Func<T, string> getFromBaseObject, int maxNumOfChars = int.MaxValue, Regex inputValidator = null) 
			: base(id,label,order,typeof(string),defaultValue,o=>getFromBaseObject((T)o)) {
			this.maxNumOfChars = maxNumOfChars;
			this.inputValidator = inputValidator;
		}
	}

    public interface ISettingsFieldPropertiesPrimitiveText
    {
        bool EmptyIsNull { get; }
    }

    public class SettingsFieldPropertiesPrimitiveText<T> : SettingsFieldPropertiesPrimitive, ISettingsFieldPropertiesPrimitiveText
    {
        public bool EmptyIsNull { get; set; } = false;
        public SettingsFieldPropertiesPrimitiveText(string id, string label, int order, string defaultValue, Func<T, string> getFromBaseObject)
            : base(id, label, order, typeof(string), defaultValue, o => getFromBaseObject((T)o))
        {}
    }

    public class SettingsFieldPropertiesPrimitiveBool<T> : SettingsFieldPropertiesPrimitive {
		public SettingsFieldPropertiesPrimitiveBool(string id, string label, int order, bool defaultValue, Func<T, bool> getFromBaseObject) 
			: base(id,label,order,typeof(bool),defaultValue,o=>getFromBaseObject((T)o)) {}
	}
	
	public class SettingsFieldPropertiesPrimitiveColor<T> : SettingsFieldPropertiesPrimitive {
		public SettingsFieldPropertiesPrimitiveColor(string id, string label, int order, Color defaultValue, Func<T, Color> getFromBaseObject) 
			: base(id,label,order,typeof(Color),defaultValue,o=>getFromBaseObject((T)o)) {}
	}
	
	public class SettingsFieldPropertiesSettings : SettingsFieldProperties {
		override public bool isSettings { get { return true; } }
		override public bool isList { get { return false; } }
		override public bool isSelectable { get { return false; } }
		public string idEnclosedSettings {get; protected set;}
        public bool allowRemoval { get; set; } = false;
        public Func<object> createNewEntry { get; set; } = null;

        public SettingsFieldPropertiesSettings(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject, string idEnclosedSettings) 
			: base(id,label,order,type,defaultValue,getFromBaseObject)
		{
			if (!typeof(ISettingsInstance).IsAssignableFrom(type))
				throw new ArgumentException("type for settings field must implement the interface \""+typeof(ISettingsInstance)+"\".");
			if (idEnclosedSettings == null)
				throw new ArgumentNullException("idEnclosedSettings");
			this.idEnclosedSettings = idEnclosedSettings;
		}
		
		override public object deepCopyValue(object obj) {
			ISettingsInstance si = obj as ISettingsInstance;
			if ( si != null ) {
				SettingsInstance siNew = new SettingsInstance();
				siNew.adopt(si);
				return siNew;
			}
			return null;
		}
	}
	
	public class SettingsFieldPropertiesSelectable : SettingsFieldProperties, ISettingsFieldPropertiesSelectSingle {
		static public Func<object, String> DefLabelProducer = o=>o != null ? ((Def)o).label.NullOrEmpty() ? ((Def)o).defName : ((Def)o).label : "NONE";
		static public Func<object, String> DefLabelProducerUnique = o=>o != null ? (((Def)o).label.NullOrEmpty() ? ((Def)o).defName : ((Def)o).label +" ("+((Def)o).defName+")") : "NONE";
		public Func<Flow, Func<object, bool>> onSelectItem { get; protected set; }
		public Func<Flow, IEnumerable<object>> selectableItems { get; protected set; }
		public Func<object, String> labelProducer { get; protected set; }
        public Func<object, Texture2D> iconProducer { get; protected set; }
        public bool isNullAllowed { get; protected set; }
		override public bool isSettings { get { return false; } }
		override public bool isList { get { return false; } }
		override public bool isSelectable { get { return true; } }
		
		public SettingsFieldPropertiesSelectable(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject, Func<Flow, IEnumerable<object>> selectableItems, Func<Flow, Func<object, bool>> onSelectItem, Func<object, String> labelProducer, bool isNullAllowed, Func<object, Texture2D> iconProducer = null) 
			: base(id,label,order,type,defaultValue,getFromBaseObject)
		{
			this.selectableItems = selectableItems;
			this.onSelectItem = onSelectItem;
			this.isNullAllowed = isNullAllowed;
			this.labelProducer = labelProducer;
            this.iconProducer = iconProducer;

            if ( isNullAllowed ) {
				this.selectableItems = flow => { IEnumerable<object> lst = selectableItems(flow); if ( !lst.Contains(null) ) { lst = new object[]{null}.Concat(lst); } return lst; };
			}
		}
	}

    public class SettingsFieldPropertiesSelectableIcon : SettingsFieldPropertiesSelectable
    {
        public SettingsFieldPropertiesSelectableIcon(string id, string label, int order, KeyValuePair<string, Texture2D>? defaultValue, Func<object, string> getFromBaseObject, string[] iconDirectories, string[] iconPaths, bool isNullAllowed = true)
            : base(id, label, order, typeof(KeyValuePair<string, Texture2D>?), defaultValue, 
                  o => GetFromString(getFromBaseObject(o)), 
                  f => Selectables(f, iconDirectories, iconPaths), 
                  f => o => OnSelectItem(f, id, (KeyValuePair<string, Texture2D>?)o), 
                  LabelProducer, 
                  isNullAllowed, 
                  IconProducer)
        {
            
        }

        static protected string LabelProducer(object kvp)
        {
            if (kvp != null)
            {
                if (kvp is string) return (string)kvp;
                else return ((KeyValuePair<string, Texture2D>)kvp).Key;
            }
            return null;
        }

        static protected Texture2D IconProducer(object kvp)
        {
            return kvp != null ? ((KeyValuePair<string, Texture2D>)kvp).Value : null;
        }

        static protected IEnumerable<object> Selectables(Flow f, string[] iconDirectories, string[] iconPaths)
        {
            //Dictionary that is uniquely filled with all icons in directories or files, assets are loaded case insensitive so we compare on lowercase
            Dictionary<string, KeyValuePair<string, Texture2D>> dic = new Dictionary<string, KeyValuePair<string, Texture2D>>();
            if (iconDirectories != null)
            {
                foreach (string s in iconDirectories)
                {
                    string directory;
                    if (!s.EndsWith("/")) directory = s + "/";
                    else directory = s;

                    //loads from one directory and filters out textures from subdirectories
                    foreach (Texture2D tex in ContentFinder<Texture2D>.GetAllInFolder(directory))
                    {
                        string path = directory + tex.name;
                        if (dic.ContainsKey(path.ToLower())) continue;
                        Texture2D texConfirm = ContentFinder<Texture2D>.Get(path, false);
                        if (texConfirm != null) dic.Add(path.ToLower(), new KeyValuePair<string, Texture2D>(path, texConfirm));
                    }
                }
            }
            if (iconPaths != null)
            {
                foreach (string iconPath in iconPaths)
                {
                    if (dic.ContainsKey(iconPath.ToLower())) continue;
                    Texture2D tex = ContentFinder<Texture2D>.Get(iconPath, false);
                    if (tex != null)
                    {
                        dic.Add(iconPath.ToLower(), new KeyValuePair<string, Texture2D>(iconPath, tex));
                    }
                }
            }
            return dic.Values.OrderBy(kvp=>kvp.Key).Cast<object>();
        }

        static protected bool OnSelectItem(Flow f, string id, KeyValuePair<string, Texture2D>? kvp)
        {
            ((ISettingsInstance)f.flowScope["edit"]).set<KeyValuePair<string, Texture2D>?>(id, kvp);
            return true;
        }

        static protected string GetFromString(string iconPath)
        {
            if (iconPath == null) return null;
            Texture2D tex = ContentFinder<Texture2D>.Get(iconPath, false);
            if (tex == null) return null;
            return iconPath;
        }

        static public bool UpdateIconpath(ref string iconPath, KeyValuePair<string,Texture2D>? kvp)
        {
            string oldPath = iconPath;
            if (!kvp.HasValue)
            {
                iconPath = null;
            }
            else
            {
                iconPath = kvp.Value.Key;
            }
            return oldPath != iconPath;
        }
    }

    public abstract class SettingsFieldPropertiesList : SettingsFieldProperties, ISettingsFieldPropertiesList {
		virtual public bool isListSettings { get { throw new NotImplementedException(); } }
		virtual public bool isListPrimitive { get { throw new NotImplementedException(); } }
		public Func<object> createNewEntry { get; protected set; }
		override public bool isSettings { get { return false; } }
		override public bool isList { get { return true; } }
		override public bool isSelectable { get { return false; } }
		public bool isRemovalAllowed { get; protected set; }
		public bool isAddingAllowed { get; protected set; }
        abstract public bool isSortable { get; }

        protected SettingsFieldPropertiesList(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject, Func<object> createNewEntry = null, bool isRemovalAllowed = false) 
			: base(id,label,order,type,defaultValue,getFromBaseObject)
		{
			this.createNewEntry = createNewEntry;
			this.isAddingAllowed = createNewEntry != null;
			this.isRemovalAllowed = isRemovalAllowed;
		}
		
		override public object deepCopyValue(object obj) { throw new NotImplementedException(); }
	}
	
	public class SettingsFieldPropertiesListPrimitive : SettingsFieldPropertiesList {
		override public bool isListSettings { get { return false; } }
		override public bool isListPrimitive { get { return true; } }
		override public bool isSelectable { get { return false; } }
		public Func<ISettingsInstance,Flow,string> listLabelProducer {get; protected set;}
        override public bool isSortable { get { return sortable; } }
        public bool sortable = true;

        public SettingsFieldPropertiesListPrimitive(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject, Func<ISettingsInstance,Flow,string> listLabelProducer, Func<object> createNewEntry = null, bool isRemovalAllowed = false)
			: base(id,label,order,type,defaultValue,getFromBaseObject,createNewEntry,isRemovalAllowed)
		{
			this.listLabelProducer = listLabelProducer;
		}
		
		override public object deepCopyValue(object obj) {
			if ( obj == null ) return null;
			IList lst = obj as IList;
			IList lstNew = (IList)Activator.CreateInstance(obj.GetType());
			foreach ( object item in lst ) lstNew.Add(item);
			return lstNew;
		}
	}
	
	public class SettingsFieldPropertiesListSettings : SettingsFieldPropertiesList {
		override public bool isListSettings { get { return true; } }
		override public bool isListPrimitive { get { return false; } }
		override public bool isSelectable { get { return false; } }
		public string idEnclosedSettings { get; protected set; }
        override public bool isSortable { get { return sortable; } }
        public bool sortable = false;

        public SettingsFieldPropertiesListSettings(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject, string idEnclosedSettings, Func<object> createNewEntry = null, bool isRemovalAllowed = false)
			: base(id,label,order,type,defaultValue,getFromBaseObject,createNewEntry,isRemovalAllowed)
		{
			if (idEnclosedSettings == null)
				throw new ArgumentNullException("idEnclosedSettings");
			this.idEnclosedSettings = idEnclosedSettings;
		}
		
		override public object deepCopyValue(object obj) {
			if ( obj == null ) return null;
			IList lst = obj as IList;
			IList lstNew = (IList)Activator.CreateInstance(obj.GetType());
			foreach ( ISettingsInstance item in lst ) {
				SettingsInstance si = new SettingsInstance();
				si.adopt(item);
				lstNew.Add(si);
			}
			return lstNew;
		}
	}

    public class SettingsFieldPropertiesListSettingsReferences : SettingsFieldPropertiesList
    {
        override public bool isListSettings { get { return true; } }
        override public bool isListPrimitive { get { return false; } }
        override public bool isSelectable { get { return false; } }
        override public bool isSortable { get { return sortable; } }
        public bool sortable = true;

        public string idEnclosedSettings { get; protected set; }
        protected Func<object> createNewBaseObjectInstance;
        public object NewBaseObjectInstance { get { if (createNewBaseObjectInstance != null) return createNewBaseObjectInstance(); return null; } }

        public SettingsFieldPropertiesListSettingsReferences(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject, string idEnclosedSettings, Func<object> createNewBaseObjectInstance = null, bool isRemovalAllowed = false)
            : base(id, label, order, type, defaultValue, getFromBaseObject, ()=> new SettingsInstance(idEnclosedSettings, createNewBaseObjectInstance(), null, null), isRemovalAllowed)
        {
            this.idEnclosedSettings = idEnclosedSettings ?? throw new ArgumentNullException("idEnclosedSettings");
            this.createNewBaseObjectInstance = createNewBaseObjectInstance;
        }

        override public object deepCopyValue(object obj)
        {
            if (obj == null) return null;
            IList lst = obj as IList;
            IList lstNew = (IList)Activator.CreateInstance(obj.GetType());
            foreach (ISettingsInstance item in lst)
            {
                SettingsInstance si = new SettingsInstance();
                si.adopt(item);
                lstNew.Add(si);
            }
            return lstNew;
        }
    }

    //TODO: SettingsField can be a selectable primitive and a selectable setting
    public class SettingsFieldPropertiesListSelectable : SettingsFieldPropertiesList {
		override public bool isListSettings { get { return false; } }
		override public bool isListPrimitive { get { return false; } }
		override public bool isSelectable { get { return true; } }
		public Func<Flow, Func<object, bool>> onSelectItem { get; protected set; }
		public Func<Flow, IEnumerable<object>> selectableItems { get; protected set; }
		public Func<object,string> labelProducer {get; protected set;}
		public Func<object,string> tooltipProducer {get; protected set;}
        override public bool isSortable { get { return sortable; } }
        public bool sortable = true;

        public SettingsFieldPropertiesListSelectable(string id, string label, int order, Type type, object defaultValue, Func<object, object> getFromBaseObject, Func<Flow, IEnumerable<object>> selectableItems, Func<Flow, Func<object, bool>> onSelectItem, Func<object,string> labelProducer, Func<object> createNewEntry = null, bool isRemovalAllowed = false, Func<object,string> tooltipProducer = null)
			: base(id,label,order,type,defaultValue,getFromBaseObject,createNewEntry,isRemovalAllowed)
		{
			this.selectableItems = selectableItems;
			this.onSelectItem = onSelectItem;
			this.labelProducer = labelProducer;
			this.tooltipProducer = tooltipProducer;
		}
		
		override public object deepCopyValue(object obj) {
			if ( obj == null ) return null;
			IList lst = obj as IList;
			IList lstNew = (IList)Activator.CreateInstance(obj.GetType());
			foreach ( object item in lst ) lstNew.Add(item);
			return lstNew;
		}
	}
}
