﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 30.10.2017
 * Time: 06:19
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Text;
using Verse;

namespace RIMMSqol.genericSettings
{
	/// <summary>
	/// Description of SettingsInstance.
	/// </summary>
	public interface ISettingsInstance
	{
		string getId();
		string getKey();
		bool getActive();
		void setActive(bool newValue);
		bool getDirty();
		void setDirty(bool newValue);
		string getLabel();
		V get<V>(string fieldId);
        void set<V>(string fieldId, V value);
        void setUnchanged<V>(string fieldId, V value);
		void markAsChanged();
		void reset();
		void attach(ISettingsInstance parentObject, ISettingsFieldProperties field);
		ISettingsInstance getParentObject();
        ISettingsFieldProperties getParentField();
        V prop<V>(string propId);
		void resolveProps(SettingsProperties properties, object baseObject);
		void resetFields(object baseObject, string[] fieldIds);
		void resetUnconfiguredFields(ISettingsInstance siOriginal = null);
        void resetFieldsToDefault();
        void markFieldAsConfigured(string fieldId, bool bubbles = true, bool triggerMarkAsChanged = true);
		void markFieldAsNotConfigured(string fieldId, bool triggerMarkAsChanged = true);
		bool isFieldConfigured(string fieldId);
		bool hasFieldConfiguredStatus();
		void initFieldConfigurationStatusHierachy(bool presetState);
        int BaseObjectReferenceKey { get; }
        object BaseObjectReference { get; }
        void OnStartEditing();
        void OnStopEditing();
    }
	
	public class SettingsInstance : ISettingsInstance, IExposable {
		protected bool active, dirty;
		protected Dictionary<string,ISettingsField> _fields;
		protected ISettingsInstance parentObject;
        protected ISettingsFieldProperties parentField;
        protected string id, baseObjectKey, label;
		protected Func<ISettingsInstance,string> dynamicLabel;
		protected Dictionary<string,object> props;
		protected Dictionary<string,bool> _isFieldConfigured;
		public bool getActive() { if ( parentObject == null ) return active; else return parentObject.getActive(); }
		public void setActive(bool newValue) { if ( parentObject == null ) active = newValue; else parentObject.setActive(newValue); }
		public bool getDirty() { if ( parentObject == null ) return dirty; else return parentObject.getDirty(); }
		public void setDirty(bool newValue) { if ( parentObject == null ) dirty = newValue; else parentObject.setDirty(newValue); }
		public string getId() { return id; }
		public string getKey() { return baseObjectKey; }
		public string getLabel() {			
			if ( label == null ) {
				SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
				if ( properties.dynamicLabel != null ) label = properties.dynamicLabel(this);
			}
			return label;
		}
		public void attach(ISettingsInstance parentObject, ISettingsFieldProperties parentField) { this.parentObject = parentObject; this.parentField = parentField; }
		public ISettingsInstance getParentObject() { return parentObject; }
        public ISettingsFieldProperties getParentField() { return parentField; }

        protected int _baseObjectReferenceKey = -1;
        public bool HasBaseObjectReference { get { return BaseObjectReferenceKey >= 0 && BaseObjectReference != null; } }
        public int BaseObjectReferenceKey { get => _baseObjectReferenceKey; set => _baseObjectReferenceKey = value; }
        public object BaseObjectReference { get; set; }

        public V prop<V>(string propId) {
			object propValue;
			if ( props.TryGetValue(propId, out propValue) ) {
				return (V)propValue;
			}
			return default(V);
		}
		
		public void resolveProps(SettingsProperties properties, object baseObject) {
			props = new Dictionary<string, object>();
			foreach ( SettingsPropertyProperties pp in properties.props ) {
				props.Add(pp.id,pp.resolver(baseObject));
			}
		}

        public SettingsInstance setProp(String propId, object propValue)
        {
            if ( props == null ) props = new Dictionary<string, object>();
            props.Add(propId, propValue);
            return this;
        }

        public SettingsInstance() {}
		
		public SettingsInstance(string id, object obj) : this(id,obj,null,null) {
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			ISettingsVisualizationSelection selection = properties.visualization as ISettingsVisualizationSelection;
			if ( selection != null ) {
				this.baseObjectKey = selection.getKeyFromBaseObject(obj);
				this.label = selection.getLabelFromBaseObject(obj);
			} else {
				this.baseObjectKey = "singleton";
				this.label = "singleton";
			}
			//resolveProps(properties, obj);
		}
		
		public SettingsInstance(string id, object obj, string baseObjectKey, string label) {
			this.id = id;
			this.baseObjectKey = baseObjectKey;
			this.label = label;
			this.active = false;
			this.dirty = false;
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			foreach ( SettingsFieldProperties p in properties.fields ) {
				fields[p.id].set(p.getFromBaseObject(obj));
				attachFieldIfPossible(p.id);
			}
			if ( properties.visualization.displayInMenu )
				initFieldConfigurationStatusHierachy(false);
			resolveProps(properties, obj);
		}
		
		public bool hasFieldConfiguredStatus() {
			return _isFieldConfigured != null;
		}

		public bool isFieldConfigured(string fieldId) {
			if ( !hasFieldConfiguredStatus() ) return true;
			bool res;
			if ( _isFieldConfigured.TryGetValue(fieldId, out res) ) {
				return res;
			}
			Log.Error("Failed to locate configured status for fieldId: "+fieldId);
			return true;
		}
		
		public void markFieldAsConfigured(string fieldId, bool bubble = true, bool triggerMarkAsChanged = true) {
			if ( hasFieldConfiguredStatus() ) {
				bool current;
				if ( !_isFieldConfigured.TryGetValue(fieldId, out current) || !current )
					_isFieldConfigured[fieldId] = true;
			}
			
			//This has to bubble up if possible so that parent settings are also marked as configured
			if ( bubble && parentObject != null ) {
				SettingsProperties settings = SettingsStorage.getSettingsProperties(parentObject.getId());
				foreach ( ISettingsFieldProperties field in settings.fields ) {
					if ( field.isSettings ) {
						ISettingsInstance child = parentObject.get<ISettingsInstance>(field.id);
						if ( child == this ) {
							parentObject.markFieldAsConfigured(field.id);
						}
					} else if ( field.isList && ((ISettingsFieldPropertiesList)field).isListSettings ) {
						IList child = parentObject.get<IList>(field.id);
						if ( child != null ) {
							foreach ( ISettingsInstance exposable in (IList)child ) {
								if ( exposable == this ) {
									parentObject.markFieldAsConfigured(field.id);
									break;
								}
							}
						}
					}
				}
			}
			
			if ( triggerMarkAsChanged ) markAsChanged();
		}
		
		public void disableFieldConfigurationStatus() {
			_isFieldConfigured = null;
		}
		
		public void initFieldConfigurationStatus(bool presetState, bool bubble = true) {
			_isFieldConfigured = new Dictionary<string, bool>();
			
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			foreach ( SettingsFieldProperties p in properties.fields ) {
				if ( !_isFieldConfigured.ContainsKey(p.id) ) {
					if ( p.id == "Version" ) {
						markFieldAsConfigured(p.id, false, false);
					} else {
						if ( presetState )
							markFieldAsConfigured(p.id, bubble, false);
						else 
							markFieldAsNotConfigured(p.id, false);
					}
				}
			}
		}
		
		public void initFieldConfigurationStatusHierachy(bool presetState) {
			SettingsProperties settings = SettingsStorage.getSettingsProperties(getId());
			//We scan for strong linked setting instances and recur the call. In this instance we initialize the configuration status if it hasn't been set yet.
			foreach ( ISettingsFieldProperties field in settings.fields ) {
				if ( field.isSettings ) {
					ISettingsInstance child = get<ISettingsInstance>(field.id);
					if ( child != null ) child.initFieldConfigurationStatusHierachy(presetState);
				} else if ( field.isList ) {
					ISettingsFieldPropertiesList lstSettings = (ISettingsFieldPropertiesList)field;
					if ( lstSettings.isListSettings && !lstSettings.isAddingAllowed && !lstSettings.isRemovalAllowed ) {
						IList child = get<IList>(field.id);
						if ( child != null ) {
							foreach ( ISettingsInstance si in (IList)child ) {
								si.initFieldConfigurationStatusHierachy(presetState);
							}
						}
					}
				}
			}
			
			if ( !hasFieldConfiguredStatus() ) {
				initFieldConfigurationStatus(presetState, false);
			}
		}
		
		public void markFieldAsNotConfigured(string fieldId, bool triggerMarkAsChanged = true) {
			if ( !hasFieldConfiguredStatus() ) return;
			bool current;
			if ( _isFieldConfigured.TryGetValue(fieldId, out current) && !current ) return;
			if ( fieldId == "Version" ) return;
			_isFieldConfigured[fieldId] = false;
			
			if ( triggerMarkAsChanged ) markAsChanged();
		}
		
		protected void updateDynamicLabel() {
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			if ( properties.dynamicLabel != null ) label = null;//properties.dynamicLabel(this);
		}
		
		protected Dictionary<string,ISettingsField> fields {
			get {
				if ( _fields == null || _fields.Count == 0 ) {
					SettingsProperties settings = SettingsStorage.getSettingsProperties(getId());
					_fields = new Dictionary<string, ISettingsField>();
					foreach ( SettingsFieldProperties p in settings.fields ) {
						ISettingsField value;
						if ( p.type == typeof(int) | p.type == typeof(float) | p.type == typeof(string) | p.type == typeof(bool) | p.type == typeof(UnityEngine.Color) ) {
							value = new PrimitiveSettingsField();
						} else if ( p.type == typeof(Def) || p.type.IsSubclassOf(typeof(Def)) ) {
							value = new DefSettingsField();
                        } else if (p is SettingsFieldPropertiesSelectableIcon) {
                            value = new IconSettingsField();
                        } else if ( typeof(IList).IsAssignableFrom(p.type) ) {
							value = new ListSettingsField();
						} else if ( typeof(ISettingsInstance).IsAssignableFrom(p.type) ) {
							value = new SettingsSettingsField();
						} else if ( p.type == typeof(Type) ) {
							value = new TypeSettingsField();
						} else {
							value = null;
							Log.Error("unknown type for settings: "+p.type);
						}
						if ( value != null ) {
							value.set(p.defaultValue);
							_fields.Add(p.id,value);
						}
					}
				}
				return _fields;
			}
		}
		
		public V get<V>(string fieldId) {
			ISettingsField field;
			if ( fields.TryGetValue(fieldId, out field) ) {
				return (V)field.get();
			}
			return default(V);
		}
        
        protected void attachFieldIfPossible(string fieldId) {
			//if a field that contains another ISettingsInstane object or a container that contains more of those is changed we have to attach those objects to this one
			ISettingsField field;
			if ( fields.TryGetValue(fieldId, out field) ) {
				//only grab the value if the field is a setting otherwise a def might not be ready to be resolved and cause problems.
				SettingsProperties settings = SettingsStorage.getSettingsProperties(getId());
				ISettingsFieldProperties fieldSettings = settings.fields.Find(f=>f.id.Equals(fieldId));
				if ( fieldSettings.isSettings || (fieldSettings.isList && ((ISettingsFieldPropertiesList)fieldSettings).isListSettings) ) {
					object value = field.get();
					if ( value != null ) {
						if ( fieldSettings.isSettings ) {
							((ISettingsInstance)value).attach(this, fieldSettings);
						} else if ( fieldSettings.isList && ((ISettingsFieldPropertiesList)fieldSettings).isListSettings ) {
							foreach ( ISettingsInstance exposable in (IList)value ) {
								exposable.attach(this, fieldSettings);
							}
						}
					}
				}
			}
		}
		
		public void resetFields(object baseObject, string[] fieldIds) {
			if ( fieldIds.NullOrEmpty() ) return;
			HashSet<string> lookup = new HashSet<string>(fieldIds);
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			if ( properties != null ) {
				foreach ( SettingsFieldProperties p in properties.fields ) {
					if ( lookup.Contains(p.id) ) {
						fields[p.id].set(p.getFromBaseObject(baseObject));
						markFieldAsNotConfigured(p.id);
						attachFieldIfPossible(p.id);
					}
				}
			}
		}
		
		public void resetFields(ISettingsInstance originalSettings, string[] fieldIds) {
			if ( fieldIds.NullOrEmpty() ) return;
			HashSet<string> lookup = new HashSet<string>(fieldIds);
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			if ( properties != null ) {
				/*MethodInfo miGetNonGeneric = this.GetType().GetMethod("get", BindingFlags.Instance|BindingFlags.Public);
				MethodInfo miSetNonGeneric = this.GetType().GetMethod("set", BindingFlags.Instance|BindingFlags.Public);*/
				foreach ( SettingsFieldProperties p in properties.fields ) {
					if ( lookup.Contains(p.id) ) {
						adoptField(this,originalSettings,p);
						//miSetNonGeneric.MakeGenericMethod(p.type).Invoke(this,new Object[]{p.id,miGetNonGeneric.MakeGenericMethod(p.type).Invoke(originalSettings,new Object[]{p.id})});
						markFieldAsNotConfigured(p.id);
						attachFieldIfPossible(p.id);
					}
				}
			}
		}
		
		public void set<V>(string fieldId, V value)
        {
			ISettingsField field;
			if ( fields.TryGetValue(fieldId, out field) ) {
				if ( field.set(value) ) {
					markFieldAsConfigured(fieldId);
					attachFieldIfPossible(fieldId);
					markAsChanged();
				}
			}
		}

        public void setUnchanged<V>(string fieldId, V value)
        {
            ISettingsField field;
            if (fields.TryGetValue(fieldId, out field))
            {
                if (field.set(value))
                {
                    attachFieldIfPossible(fieldId);
                }
            }
        }

        static protected Dictionary<Type, Dictionary<Type, MethodInfo>> genericFieldGetters = new Dictionary<Type, Dictionary<Type, MethodInfo>>();
		static protected Dictionary<Type, Dictionary<Type, MethodInfo>> genericFieldSetters = new Dictionary<Type, Dictionary<Type, MethodInfo>>();
		
		static protected void adoptField(ISettingsInstance si, ISettingsInstance source, SettingsFieldProperties field) {
			try {
				Type t = si.GetType();
				Dictionary<Type, MethodInfo> getters;
				if ( !genericFieldGetters.TryGetValue(t, out getters) ) {
					getters = new Dictionary<Type, MethodInfo>();
					genericFieldGetters.Add(t,getters);
				}
				Dictionary<Type, MethodInfo> setters;
				if ( !genericFieldSetters.TryGetValue(t, out setters) ) {
					setters = new Dictionary<Type, MethodInfo>();
					genericFieldSetters.Add(t,setters);
				}
				MethodInfo getter;
				if ( !getters.TryGetValue(field.type, out getter) ) {
					getter = t.GetMethod("get", BindingFlags.Instance|BindingFlags.Public).MakeGenericMethod(field.type);
					getters.Add(field.type, getter);
			    }
				MethodInfo setter;
				if ( !setters.TryGetValue(field.type, out setter) ) {
					setter = t.GetMethod("set", BindingFlags.Instance|BindingFlags.Public).MakeGenericMethod(field.type);
					setters.Add(field.type, setter);
			    }
                object o = getter.Invoke(source, new Object[] { field.id });
                setter.Invoke(si,new Object[]{field.id,field.deepCopyValue(getter.Invoke(source,new Object[]{field.id}))});
			} catch (Exception ex) {
				Log.Error("Failed to adopt field: "+field.id+" on "+si.getKey()+" of "+si.getId()+" with message "+ex.Message+"\n"+ex.StackTrace);
			}
		}
		
		public virtual void adopt(ISettingsInstance source) {
			if ( this == source ) return;
			id = source.getId();
			baseObjectKey = source.getKey();
			label = source.getLabel();
            BaseObjectReferenceKey = source.BaseObjectReferenceKey;
            BaseObjectReference = source.BaseObjectReference;
            fields.Clear();
			SettingsProperties settings = SettingsStorage.getSettingsProperties(getId());
			foreach ( SettingsFieldProperties field in settings.fields ) {
				adoptField(this,source,field);
			}
			if ( source.hasFieldConfiguredStatus() ) {
				initFieldConfigurationStatus(false, false);
				foreach ( SettingsFieldProperties field in settings.fields ) {
					if ( source.isFieldConfigured(field.id) ) {
						markFieldAsConfigured(field.id, false, false);
					}
				}
			} else disableFieldConfigurationStatus();
			active = source.getActive();
			dirty = source.getDirty();
			props = new Dictionary<string, object>();
			foreach ( SettingsPropertyProperties pp in settings.props ) {
				props.Add(pp.id,source.prop<object>(pp.id));
			}
		}
		
		public virtual void markAsChanged() {
			updateDynamicLabel();
			if ( parentObject == null ) {
				active = true;
				dirty = true;
			} else {
				parentObject.markAsChanged();
			}
		}
		
		public virtual void reset() { 
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			ISettingsInstance originalSettings;
			if ( properties.overriddenSettings.TryGetValue(getKey(), out originalSettings) ) {
				adopt(originalSettings);
			} else {
				adopt(new SettingsInstance(id,((ISettingsVisualizationSelection)properties.visualization).getBaseObjectFromKey(getKey())));
			}
			active = false;
			dirty = false;
		}

        protected bool createdOverrideForEditing = false;

        public virtual void OnStartEditing()
        {
            //in order for references to be available to the editor they are exposed in the form of an override entry before the settings are actually merged
            SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
            if (!properties.visualization.displayInMenu) return;
            if (!properties.overriddenSettings.ContainsKey(getKey()))
            {
                createdOverrideForEditing = SettingsStorage.AddOverrideSetting(properties, getKey());
            }
            else createdOverrideForEditing = false;
        }

        public virtual void OnStopEditing()
        {
            //if an override entry was created for only exposing references it is deleted after editing if this instance is not active
            if (createdOverrideForEditing)
            {
                try
                {
                    if (!getActive())
                    {
                        SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
                        if (!properties.visualization.displayInMenu) return;
                        properties.overriddenSettings.Remove(getKey());
                    }
                } finally { createdOverrideForEditing = false; }
            }
        }

        protected virtual void OnSerialize() {
			Scribe_Values.Look<string>(ref baseObjectKey, "baseObjectKey", null);
			Scribe_Values.Look<string>(ref label, "label", null);
            //default is -2 so that we can distinguish between not present(-2 = old entries) and not set(-1 = new instance that uses no reference)
            Scribe_Values.Look<int>(ref _baseObjectReferenceKey, "baseObjectReferenceKey", -2);
        }
		
		protected virtual void OnBeforeSerialize() {}
		
		protected virtual void OnAfterDeserialize() {
			foreach ( string fieldId in fields.Keys ) {
				attachFieldIfPossible(fieldId);
			}
			active = true;
			dirty = true;
		}

        public static bool exposeUnconfiguredFields = false;
        public static HashSet<string> exposeOverriddenValuesInSettingsId = new HashSet<string>();

        public void ExposeData()
		{
			if ( Scribe.mode == LoadSaveMode.LoadingVars || Scribe.mode == LoadSaveMode.Saving ) {
				Scribe_Values.Look<string>(ref id, "id", null);
				SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
				
				//settings inside a parent object that cannot be matched with original values(e.g. mutatable lists) cannot be configured.
				if ( Scribe.mode == LoadSaveMode.LoadingVars ) {
					string str = null;
					Scribe_Values.Look<string>(ref str, "isConfigured");
					if ( str == null ) {
						disableFieldConfigurationStatus();
					} else {
						initFieldConfigurationStatus(true);
						Dictionary<string,bool> read = new Dictionary<string, bool>();
						string[] parts = str.Split(';');
						for ( int i = 0; i < parts.Length-1; i+=2 ) {
							read.Add(parts[i],parts[i+1]=="t");
						}
						foreach ( SettingsFieldProperties p in properties.fields ) {
							if ( read.ContainsKey(p.id) ) {
								if ( !read[p.id] ) {
									markFieldAsNotConfigured(p.id, false);
								}
							} else {
								//All fields that are not listed in isConfigured should be markes as not configured since they are new properties.
								markFieldAsNotConfigured(p.id, false);
							}
						}
					}
				} else {
					if ( !hasFieldConfiguredStatus() ) {
						string str = null;
						Scribe_Values.Look<string>(ref str, "isConfigured");
					} else {
                        bool useOverriddenValues = exposeOverriddenValuesInSettingsId.Contains(properties.id);
                        StringBuilder sb = new StringBuilder();
						foreach ( KeyValuePair<string,bool> p in _isFieldConfigured ) {
							sb.Append(p.Key).Append(';').Append(p.Value && !useOverriddenValues ? 't':'f').Append(';');
						}
						foreach ( SettingsFieldProperties p in properties.fields ) {
							if ( !_isFieldConfigured.ContainsKey(p.id) ) {
								sb.Append(p.id).Append(';').Append('f').Append(';');
							}
						}
						string str = sb.ToString();
						Scribe_Values.Look<string>(ref str, "isConfigured");
					}
				}
				
				if (Scribe.mode == LoadSaveMode.Saving) {
					OnBeforeSerialize();
					foreach ( SettingsFieldProperties p in properties.fields ) {
						ISettingsField field; if ( (exposeUnconfiguredFields || isFieldConfigured(p.id)) && fields.TryGetValue(p.id,out field) ) field.OnBeforeSerialize(p);
					}
				}
				OnSerialize();
				foreach ( SettingsFieldProperties p in properties.fields ) {
					ISettingsField field; if ((exposeUnconfiguredFields || isFieldConfigured(p.id)) && fields.TryGetValue(p.id, out field))
                    {
                        if (exposeOverriddenValuesInSettingsId.Contains(properties.id) && properties.visualization.displayInMenu) getOverriddenField(properties, p.id, ref field);
                        field.ExposeData(p);
                    }
				}
				
				if (Scribe.mode != LoadSaveMode.Saving) {
					OnAfterDeserialize();
					foreach ( SettingsFieldProperties p in properties.fields ) {
						ISettingsField field; if ((exposeUnconfiguredFields || isFieldConfigured(p.id)) && fields.TryGetValue(p.id,out field) ) field.OnAfterDeserialize(p);
					}
				}
				
				//after loading if it is a top level property with selection defined we can trigger an initialization for the configuration state on all stages where it is missing
				if ( properties.visualization.displaySelection ) initFieldConfigurationStatusHierachy(true);
			}
		}

        protected void getOverriddenField(SettingsProperties properties, string fieldId, ref ISettingsField field)
        {
            //non root settings may not have a key
            if (getKey() == null) return;
            if (properties.overriddenSettings.TryGetValue(getKey(), out ISettingsInstance siOriginal))
            {
                ((SettingsInstance)siOriginal).fields.TryGetValue(fieldId, out field);
            }
        }

        public void resetFieldsToDefault()
        {
            SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
            foreach (SettingsFieldProperties p in properties.fields)
            {
                fields[p.id].set(p.defaultValue);

                ISettingsField field;
                if (fields.TryGetValue(p.id, out field))
                {
                    if (p.isSettings)
                    {
                        SettingsInstance si = get<SettingsInstance>(p.id);
                        if (si != null)
                        {
                            si.resetFieldsToDefault();
                        }
                    }
                    else if (p.isList)
                    {
                        ISettingsFieldPropertiesList lstSettings = (ISettingsFieldPropertiesList)p;
                        if (lstSettings.isListSettings)
                        {
                            IList lstSi = get<IList>(p.id);
                            if (lstSi != null && !lstSettings.isAddingAllowed && !lstSettings.isRemovalAllowed)
                            {
                                for (int i = 0; i < lstSi.Count; i++)
                                {
                                    if (lstSi[i] != null) ((ISettingsInstance)lstSi[i]).resetFieldsToDefault();
                                }
                            }
                        }
                    }
                }
            }
        }
		
		public void resetUnconfiguredFields(ISettingsInstance siOriginal = null) {
			if ( !hasFieldConfiguredStatus() ) return;
			SettingsProperties properties = SettingsStorage.getSettingsProperties(getId());
			ISettingsVisualizationSelection selection;
			if ( siOriginal == null && (selection = properties.visualization as ISettingsVisualizationSelection) != null ) {
				if ( !properties.overriddenSettings.TryGetValue(getKey(), out siOriginal) ) {
					siOriginal = new SettingsInstance(getId(),selection.getBaseObjectFromKey(getKey()), getKey(), getLabel());
				} else {
					//an overridden entry should be deep copied to avoid polluting the origin by resetting and then changing.
				}
			}
            else if ( siOriginal == null && (properties.visualization is ISettingsVisualizationSingletonSelection) )
            {
                if (!properties.overriddenSettings.TryGetValue(getKey(), out siOriginal)) {
                    siOriginal = new SettingsInstance(getId(), ((ISettingsVisualizationSingletonSelection)properties.visualization).getBaseObject(), getKey(), getLabel());
                }
            }
			if ( siOriginal != null ) {
				List<string> unconfiguredFields = new List<string>();
				foreach ( SettingsFieldProperties p in properties.fields ) {
					if ( !isFieldConfigured(p.id) ) unconfiguredFields.Add(p.id);
				}
				//Log.Message("Resetting fields: "+unconfiguredFields.ToCommaList());
				resetFields(siOriginal, unconfiguredFields.ToArray());
				//We cannot reset items in a list since there is no hard link that identifies and matches items between the original list and the new list
				//However if it is a settings field the link is available
				//TODO Lists that do not allow adding or removing would also be linked 
				foreach ( SettingsFieldProperties p in properties.fields ) {
					if ( isFieldConfigured(p.id) ) {
						ISettingsField field;
						if ( fields.TryGetValue(p.id,out field) ) {
							if ( p.isSettings ) {
								SettingsInstance si = get<SettingsInstance>(p.id);
								if ( si != null && si.hasFieldConfiguredStatus() ) {
									SettingsInstance siChildOriginal = siOriginal.get<SettingsInstance>(p.id);
									if ( si != null && siChildOriginal != null ) {
										si.resetUnconfiguredFields(siChildOriginal);
									} else if ( siChildOriginal != null ) {
										set<SettingsInstance>(p.id, siChildOriginal);
									}
								}
							} else if ( p.isList ) {
								ISettingsFieldPropertiesList lstSettings = (ISettingsFieldPropertiesList)p;
								if ( lstSettings.isListSettings && !lstSettings.isAddingAllowed && !lstSettings.isRemovalAllowed ) {
									IList lstSi = get<IList>(p.id);
									IList lstSiOriginal = siOriginal.get<IList>(p.id);
									if ( lstSi != null && lstSiOriginal != null ) {
										//follows the rules of overwriting, a change in count or positioning is ignored
										for ( int i = 0; i < lstSi.Count && i < lstSiOriginal.Count; i++ ) {
											if ( lstSi[i] != null && lstSiOriginal[i] != null ) ((ISettingsInstance)lstSi[i]).resetUnconfiguredFields((ISettingsInstance)lstSiOriginal[i]);
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
}
