﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 31.10.2017
 * Time: 00:12
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Xml;
using Verse;

namespace RIMMSqol.genericSettings
{
	/// <summary>
	/// Description of SettingsStorage.
	/// </summary>
	public static class SettingsStorage
	{
		private static Dictionary<string,SettingsProperties> settingsProperties = new Dictionary<string, SettingsProperties>();
		private static Dictionary<string,List<SettingsInstance>> settingsInstances = new Dictionary<string, List<SettingsInstance>>();
		
		public static SettingsProperties getSettingsProperties(string id) {
			SettingsProperties settings;
			if ( settingsProperties.TryGetValue(id,out settings) ) return settings;
			return null;
		}
		
		public static SettingsProperties getOrInitSettingsProperties(string id) {
			SettingsProperties settings;
			if ( !settingsProperties.TryGetValue(id,out settings) ) {
				settings = new SettingsProperties(id,settingsProperties.Count);
				settingsProperties[id] = settings;
			}
			return settings;
		}
		
		public static bool removeSettingsProperties(string id) {
			return settingsProperties.Remove(id);
		}
		
		public static List<SettingsProperties> getAllSettingsProperties() {
			return settingsProperties.Values.OrderBy(p=>p.order).ToList();
		}
		
		public static void ExposeData() {
			foreach ( SettingsProperties props in settingsProperties.Values ) {
				if ( !props.mergers.NullOrEmpty() ) {
					List<SettingsInstance> lst;
					if ( !settingsInstances.TryGetValue(props.id, out lst) ) {
						lst = null;
					}
					if (Scribe.mode == LoadSaveMode.Saving) {
						lst.RemoveAll(si=>!si.getActive());
					}
					
					try {
						Scribe_Collections.Look<SettingsInstance>(ref lst,props.id);
					} catch (Exception ex) {
						Log.Error(ex.Message+" : Failed to load RimMsQol data for "+props.id+" entries. Using defaults.");
					}
					
					if (Scribe.mode == LoadSaveMode.PostLoadInit) { 
						if ( lst == null ) lst = new List<SettingsInstance>(); 
					}
					settingsInstances[props.id] = lst;
				}
			}
		}

        public static void ExposeDataFull()
        {
            foreach (SettingsProperties props in settingsProperties.Values)
            {
                if (!props.visualization.displayInMenu) continue;
                List<SettingsInstance> lst = null;
                ISettingsVisualizationSelection visualization = props.visualization as ISettingsVisualizationSelection;
                ISettingsVisualizationSingletonSelection visualizationSingleton = props.visualization as ISettingsVisualizationSingletonSelection;
                if (visualization != null)
                {
                    lst = generateListForSelection(props)?.Cast<SettingsInstance>().ToList();
                }
                else if (visualizationSingleton != null)
                {
                    lst = new List<SettingsInstance>();
                    lst.Add((SettingsInstance)generateSingletonObject(props));
                }
                
                try
                {
                    Scribe_Collections.Look<SettingsInstance>(ref lst, props.id);
                }
                catch (Exception ex)
                {
                    Log.Error(ex.Message + " : Failed to load RimMsQol data for " + props.id + " entries. Using defaults.");
                }
            }
        }

        public static List<SettingsInstance> GetInstances(SettingsProperties props)
        {
            if (settingsInstances.TryGetValue(props.id, out List<SettingsInstance> lst))
            {
                return lst;
            }
            return new List<SettingsInstance>();
        }

        public static string BuildConfigFile(Mod mod, List<MergeFile> mergeFiles)
        {
            if (mod == null) mod = QOLMod.getModHandle();
            string outputDirectory = Path.Combine(GenFilePaths.ConfigFolderPath,
                GenText.SanitizeFilename(string.Format("Mod_{0}_{1}", mod.Content.FolderName, mod.GetType().Name)));
            //Spawn a file that contains the base xml structure for the merging
            string filePath = Path.Combine(outputDirectory, GenText.SanitizeFilename(string.Format("Mod_{0}_{1}_Initial.xml", mod.Content.FolderName,
                mod.GetType().Name)));
            Scribe.saver.InitSaving(filePath, "SettingsBlock");
            ModSettings settingsFull = new QOLModSettingsFull(QOLMod.getSettings());

            try
            {
                //based on the requirements of the merge files.
                SettingsInstance.exposeUnconfiguredFields = mergeFiles.Any(mf => mf.requiresAllProperties);
                //ability to choose whether the current settings are used or the overridden ones
                mergeFiles.Aggregate(SettingsInstance.exposeOverriddenValuesInSettingsId, (set, mf) => { if ( !mf.resetCategories.NullOrEmpty() ) set.AddRange(mf.resetCategories); return set; });
                Scribe_Deep.Look<ModSettings>(ref settingsFull, "ModSettings", Array.Empty<object>());
            }
            catch(Exception ex)
            {
                Log.Error(ex.Message + "\n" + ex.StackTrace);
                return null;
            }
            finally
            {
                SettingsInstance.exposeUnconfiguredFields = false;
                if (SettingsInstance.exposeOverriddenValuesInSettingsId != null) SettingsInstance.exposeOverriddenValuesInSettingsId.Clear();
                Scribe.saver.FinalizeSaving();
            }
            XmlDocument doc = new XmlDocument();
            doc.Load(filePath);
            List<PatchOperation> patchOperations = new List<PatchOperation>();
            mergeFiles.Aggregate(patchOperations, (lst, mf) => { if (!mf.patches.NullOrEmpty()) lst.AddRange(mf.patches); return lst; });
            //Cleaning up the leftover properties at the end instead of removing them early. Allows filtering and processing until the patching is finished.
            patchOperations.Add(new PatchOperation_QoLRemoveUnconfiguredProperties("/SettingsBlock/ModSettings/*/*"));
            //Operation that rewrites the class attribute on the root node
            PatchOperationAttributeSet poChangeClass = new PatchOperationAttributeSet();
            typeof(PatchOperationAttributeSet).GetField("xpath", BindingFlags.Instance | BindingFlags.NonPublic).SetValue(poChangeClass, "/SettingsBlock/ModSettings");
            typeof(PatchOperationAttributeSet).GetField("attribute", BindingFlags.Instance | BindingFlags.NonPublic).SetValue(poChangeClass, "Class");
            typeof(PatchOperationAttributeSet).GetField("value", BindingFlags.Instance | BindingFlags.NonPublic).SetValue(poChangeClass, typeof(QOLModSettings).FullName);
            patchOperations.Add(poChangeClass);
            
            foreach (PatchOperation po in patchOperations)
            {
                po.Apply(doc);
            }

            string outputFile = Path.Combine(outputDirectory, GenText.SanitizeFilename(string.Format("Mod_{0}_{1}.xml", mod.Content.FolderName,
                mod.GetType().Name)));
            doc.Save(outputFile);

            return outputFile;
        }

        public static void initializeSettingsInstances() {
			foreach ( SettingsProperties props in settingsProperties.Values ) {
				if ( !props.mergers.NullOrEmpty() ) {
					if ( !settingsInstances.ContainsKey(props.id) ) {
						settingsInstances[props.id] = new List<SettingsInstance>();
					}
				}
			}
		}

        public static object getBaseObject(ISettingsInstance si)
        {
            if (si != null)
            {
                SettingsProperties props = getSettingsProperties(si.getId());
                if (props != null)
                {
                    if (props.visualization is ISettingsVisualizationSelection)
                    {
                        return ((ISettingsVisualizationSelection)props.visualization).getBaseObjectFromKey(si.getKey());
                    }
                    else if (props.visualization is ISettingsVisualizationSingletonSelection) 
                    {
                        return ((ISettingsVisualizationSingletonSelection)props.visualization).getBaseObject();
                    }
                }
            }
            return null;
        }

        public static Dictionary<int, object> getBaseObjectReferences(ISettingsInstance listOwner, string listFieldId)
        {
            SettingsProperties props;
            Dictionary<int, object> baseObjectReferences = new Dictionary<int, object>();

            ISettingsInstance root = listOwner;
            List<KeyValuePair<ISettingsFieldProperties,int>> lstPath = new List<KeyValuePair<ISettingsFieldProperties, int>>();
            while (true)
            {
                if (root.getParentObject() == null) break;
                props = SettingsStorage.getSettingsProperties(root.getId());
                //possible that we need to add the list index as BaseObjectReferenceKey if -2
                lstPath.Add(new KeyValuePair<ISettingsFieldProperties, int>(root.getParentField(), root.BaseObjectReferenceKey));
                root = root.getParentObject();
            }

            props = SettingsStorage.getSettingsProperties(root.getId());
            ISettingsInstance overriddenSetting;
            if (props.overriddenSettings.TryGetValue(root.getKey(), out overriddenSetting))
            {
                lstPath.Reverse();
                foreach (KeyValuePair<ISettingsFieldProperties, int> kvp in lstPath)
                {
                    bool isFound = false;
                    ISettingsFieldProperties fieldSettings = kvp.Key;
                    if (fieldSettings.isSettings)
                    {
                        overriddenSetting = overriddenSetting.get<ISettingsInstance>(fieldSettings.id);
                        isFound = true;//creates no new entries later when adding reference support the BaseObjectReferenceKey needs to be checked as well
                    }
                    else if (fieldSettings.isList && ((ISettingsFieldPropertiesList)fieldSettings).isListSettings)
                    {
                        IList lst = overriddenSetting.get<IList>(fieldSettings.id);
                        foreach (ISettingsInstance instance in lst)
                        {
                            //might need support for -2
                            if (instance.BaseObjectReferenceKey == kvp.Value)
                            {
                                overriddenSetting = instance;
                                isFound = true;
                                break;
                            }
                        }
                    }
                    if (!isFound)
                    {
                        overriddenSetting = null;
                        break;
                    }
                }

                if (overriddenSetting != null)
                {
                    List<SettingsInstance> overriddenStages = overriddenSetting.get<List<SettingsInstance>>(listFieldId);
                    foreach (SettingsInstance os in overriddenStages)
                    {
                        if (os.HasBaseObjectReference)
                        {
                            baseObjectReferences[os.BaseObjectReferenceKey] = os.BaseObjectReference;
                        }
                    }
                }
            }
            else
            {
                Log.Error("Missing override setting when resolving base references!");
            }
            return baseObjectReferences;
        }

        public static List<T> getInstancesForMerging<T>(ISettingsInstance listOwner, string listFieldId, List<T> liveList)
        {
            SettingsProperties props = SettingsStorage.getSettingsProperties(listOwner.getId());
            List<SettingsInstance> configured = listOwner.get<List<SettingsInstance>>(listFieldId);
            if (configured == null) return null;

            //backwards compatibility
            for (int i = 0; i < configured.Count; i++)
            {
                SettingsInstance si = configured[i];
                if ( si.BaseObjectReferenceKey == -2 )
                {
                    si.BaseObjectReferenceKey = i;
                }
            }
            
            Dictionary<int, object> references = getBaseObjectReferences(listOwner, listFieldId);
            //cast to a version that manages the reference lists
            SettingsFieldPropertiesListSettingsReferences field = props.fields.FirstOrDefault(f => f.id == listFieldId) as SettingsFieldPropertiesListSettingsReferences;

            List<T> instances = new List<T>();
            foreach (SettingsInstance si in configured)
            {
                if ( si.BaseObjectReferenceKey >= 0 )
                {
                    object reference;
                    if ( references.TryGetValue(si.BaseObjectReferenceKey, out reference) )
                    {
                        instances.Add((T)reference);
                        continue;
                    }
                }
                instances.Add((T)field.NewBaseObjectInstance);
            }
            return instances;
        }

        /// <summary>
        /// Returns all currently unpicked choices from new entries or references.
        /// </summary>
        /// <returns></returns>
        public static List<SettingsInstance> getSettingsInstancesForSelection(ISettingsInstance settings, SettingsFieldPropertiesListSettingsReferences fieldConfiguration)
        {
            string fieldId = fieldConfiguration.id;
            SettingsProperties props = SettingsStorage.getSettingsProperties(settings.getId());
            List<SettingsInstance> configured = settings.get<List<SettingsInstance>>(fieldId);
            Dictionary<int, object> references = getBaseObjectReferences(settings, fieldId);
            references.RemoveAll(kvp => configured.Any(e => e.BaseObjectReferenceKey == kvp.Key));

            List<SettingsInstance> choices = new List<SettingsInstance>();
            SettingsInstance newItem = (SettingsInstance)fieldConfiguration.createNewEntry();
            if (newItem != null) choices.Add(newItem);
            choices.AddRange(references.Select(kvp=>new SettingsInstance(fieldConfiguration.idEnclosedSettings,kvp.Value, null, null) { BaseObjectReferenceKey = kvp.Key, BaseObjectReference = kvp.Value }));

            return choices;
        }

        public static bool AddOverrideSetting(SettingsProperties props, string baseObjectKey)
        {
            try
            {
                Dictionary<string, ISettingsInstance> overriddenSettings = props.overriddenSettings;
                SettingsInstance si;
                if (props.visualization.displaySelection)
                {
                    si = new SettingsInstance(props.id, ((ISettingsVisualizationSelection)props.visualization).getBaseObjectFromKey(baseObjectKey));
                }
                else
                {
                    si = new SettingsInstance(props.id, ((ISettingsVisualizationSingletonSelection)props.visualization).getBaseObject());
                }
                //overriding IExposed objects requires the use of default values for the overridden instance otherwise we reset to 
                //the initial save state and not an original one.
                if (props.ResetToDefaults)
                {
                    si.resetFieldsToDefault();
                }
                overriddenSettings.Add(baseObjectKey, si);
            }
            catch (Exception ex)
            {
                Log.Warning("RIMMSQoL setting \"" + props.id + "\".\"" + baseObjectKey + "\" failed to find overriden object. Deactivating entry." + "\n" + ex.Message + "\n" + ex.StackTrace);
                return false;
            }
            return true;
        }

        public static void ApplySettings(bool reload = false) {
			foreach ( SettingsProperties props in settingsProperties.Values ) {
				if ( !props.mergers.NullOrEmpty() ) {
					List<SettingsInstance> lstOriginal;
					if ( !settingsInstances.TryGetValue(props.id,out lstOriginal) ) {
						lstOriginal = new List<SettingsInstance>();
					}
					Dictionary<string,ISettingsInstance> overriddenSettings = props.overriddenSettings;

					List<SettingsInstance> lst = new List<SettingsInstance>(lstOriginal);
					//all overridden settings must be present so we find the ones that we have to reset, not the case if the list got purged from non active settings.
					foreach ( ISettingsInstance setting in overriddenSettings.Values ) {
						if ( lst.Find(s=>s.getKey().Equals(setting.getKey())) == null ) {
							lst.Add((SettingsInstance)setting);
						}
					}
					
					foreach ( SettingsInstance setting in lst ) {
                        bool removeOverriddenSetting = false;
                        if ( setting.getActive() && (setting.getDirty() || reload) ) {
							if ( !overriddenSettings.ContainsKey(setting.getKey()) ) {
								//before we apply a new value for a def for the first time we save the original data
                                if ( !AddOverrideSetting(props, setting.getKey()) )
                                {
                                    setting.setActive(false);
                                    lstOriginal.Remove(setting);
                                    continue;
                                }
							}
						} else if ( !setting.getActive() && overriddenSettings.ContainsKey(setting.getKey()) ) {
							//no longer active but previously overridden
							setting.adopt(overriddenSettings[setting.getKey()]);
							setting.setDirty(true);
                            removeOverriddenSetting = true;
                        } else {
							//not an active setting that needs to be merged, nor a reset setting
							continue;
						}
						
						bool isValid = true;
						try {
							foreach ( Func<ISettingsInstance, bool> validator in props.validators ) {
								if ( validator != null ) isValid = isValid && validator(setting);
							}
						} catch {
							Log.Warning("Validation for configured def setting failed.");
							isValid = false;
						}
						if ( isValid ) {
							//only when using active settings do we want to discard unconfigured fields
							//settings that were already reset must always be treated as configured.
							if ( setting.getActive() ) setting.resetUnconfiguredFields();
							foreach ( Action<ISettingsInstance> merger in props.mergers ) {
								try {
									merger(setting);
								} catch (Exception ex) {									
									setting.setActive(false);
									lstOriginal.Remove(setting);
									Log.Error("RIMMSQoL setting \""+props.id+"\".\""+setting.getKey()+"\" failed to merge(possibly due to changes in the mod list). Deactivating entry but residual changes might already be present.\n"+ex.Message+"\n"+ex.StackTrace);
								}
							}
						} else {
							setting.setActive(false);
							lstOriginal.Remove(setting);
							Log.Message("RIMMSQoL setting \""+props.id+"\".\""+setting.getKey()+"\" was marked as invalid. Deactivating entry.");
						}
						setting.setDirty(false);
                        if (removeOverriddenSetting)
                        {
                            overriddenSettings.Remove(setting.getKey());
                        }
                    }
				}
			}
		}
		
		public static List<ISettingsInstance> generateListForSelection(SettingsProperties props) {
			if ( props.visualization.displaySelection ) {
				List<SettingsInstance> lst;
				if ( settingsInstances.TryGetValue(props.id,out lst) ) {
					ISettingsVisualizationSelection visualProps = ((ISettingsVisualizationSelection)props.visualization);
					List<object> allBaseObjects = visualProps.getBaseObjectsForSelection();
					foreach ( object baseObject in allBaseObjects.Where(o=>lst.Find(si=>si.getKey().Equals(visualProps.getKeyFromBaseObject(o)))==null) ) {
						string key;
						try {
							key = visualProps.getKeyFromBaseObject(baseObject);
						} catch {
							Log.Warning("Failed to load key from database object \""+baseObject+"\".");
							continue;
						}
						if ( !lst.Any(si=>si.getKey().Equals(key)) ) {
							try {
								lst.Add(new SettingsInstance(props.id,baseObject));
							} catch (Exception e) {
								Log.Warning("Failed to load database object for key \""+key+"\" into setting \""+props.id+"\".\n"+ e.Message + "\n" + e.StackTrace);
							}
						}
					}
					//lst.AddRange(allBaseObjects.Where(o=>lst.Find(si=>si.getKey().Equals(visualProps.getKeyFromBaseObject(o)))==null).Select(o=>new SettingsInstance(props.id,o)));
					
					lst = lst.OrderByDescending(si=>si.getActive()).ThenBy(si=>si.getLabel()).ToList();
					return lst.Cast<ISettingsInstance>().ToList();
				}
			}
			return null;
		}
		
		public static ISettingsInstance generateSingletonObject(SettingsProperties props) {
			ISettingsVisualizationSingletonSelection visuals = props.visualization as ISettingsVisualizationSingletonSelection;
			if ( visuals != null ) {
				List<SettingsInstance> lst;
				if ( settingsInstances.TryGetValue(props.id,out lst) ) {
					//at some point it was possible to have multiple copies in the list this just corrects that state.
					while ( lst.Count > 1 ) { lst.RemoveAt(0); }
					if ( lst.Any() ) {
						return lst.Last();
					} else {
						try {
							SettingsInstance instance = new SettingsInstance(props.id,visuals.getBaseObject());
							lst.Add(instance);
							return instance;
						} catch {
							Log.Warning("Failed to load singleton object \""+visuals.getBaseObject()+"\" into setting \""+props.id+"\"");
						}
					}
				}
			}
			return null;
		}
	}
}
