﻿using RimWorld;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UnityEngine;
using Verse;

namespace RIMMSqol.performance
{
    public class QOLDefModExtension_RenderingParameters : DefModExtension
    {
        //Pure struct without reference types allows fastest access of content inside array.
        public struct RenderingParameters
        {
            public Vector2 offset;
            public Vector2 scaling;
        }

        private RenderingParameters[] lookup;
        private QOLDefModExtension_RenderingParametersConfig config;
        public QOLDefModExtension_RenderingParametersConfig Config { get => config; protected set => config = value; }

        public QOLDefModExtension_RenderingParameters()
        {
            Config = new QOLDefModExtension_RenderingParametersConfig();
        }

        //we use references to avoid any copying of structs, it also allows write access to the cached data but probably not of any use.
        public ref Vector2 GetOffset(Rot4 facing, BodyTypeDef bodyType)
        {
            return ref lookup[bodyType.index * 4 + facing.AsInt].offset;
        }

        public ref Vector2 GetScaling(Rot4 facing, BodyTypeDef bodyType)
        {
            return ref lookup[bodyType.index * 4 + facing.AsInt].scaling;
        }

        public ref RenderingParameters GetParameters(Rot4 facing, BodyTypeDef bodyType)
        {
            return ref lookup[bodyType.index * 4 + facing.AsInt];
        }

        public Matrix4x4 BuildMatrix(Vector3 loc, Quaternion quat, Rot4 facing, BodyTypeDef bodyType)
        {
            RenderingParameters p = GetParameters(facing, bodyType);
            Matrix4x4 m = Matrix4x4.Translate(loc);
            if (quat != Quaternion.identity) m *= Matrix4x4.Rotate(quat);

            return m * Matrix4x4.Translate(new Vector3(p.offset.x, 0f, p.offset.y)) * Matrix4x4.Scale(new Vector3(p.scaling.x, 1f, p.scaling.y));
        }

        public Matrix4x4 BuildMatrix(Vector3 loc, Quaternion quat, Rot4 facing, BodyTypeDef bodyType, Vector2 meshBounds)
        {
            RenderingParameters p = GetParameters(facing, bodyType);
            Matrix4x4 m = Matrix4x4.Translate(loc);
            if (quat != Quaternion.identity) m *= Matrix4x4.Rotate(quat);

            return m * Matrix4x4.Translate(new Vector3(p.offset.x * meshBounds.x, 0f, p.offset.y * meshBounds.y)) * Matrix4x4.Scale(new Vector3(p.scaling.x, 1f, p.scaling.y));
        }

        public void Reload()
        {
            //we have 4 directions for each bodyType
            int count = DefDatabase<BodyTypeDef>.DefCount;
            int maxIndex = DefDatabase<BodyTypeDef>.AllDefsListForReading.Max(bt => bt.index);
            lookup = new RenderingParameters[(maxIndex + 1) * 4];
            //initializing the scaling with (1,1)
            for (int i = 0; i < lookup.Length; i++)
            {
                lookup[i].scaling = Vector2.one;
            }
            if ( count <= maxIndex )
            {
                Log.Warning("Max Index is larger than required for def count on body types. This will result in wasted memory for cache mapping. Count: "+count+" Max-Index: "+maxIndex);
            }
            else if ( count - 1 > maxIndex  )
            {
                Log.Error("Max Index is too small for def count on body types. Mapping body types is not possible! Count: " + count + " Max-Index: " + maxIndex);
                return;
            }

            foreach (BodyTypeDef bt in DefDatabase<BodyTypeDef>.AllDefsListForReading)
            {
                int startIndex = bt.index * 4;
                foreach (QOLDefModExtension_RenderingParametersConfig_Entry e in Config.entries)
                {
                    for ( int facingOffset = 0; facingOffset < 4; facingOffset++ )
                    {
                        ref RenderingParameters rp = ref lookup[startIndex + facingOffset];
                        if (e.matches(facingOffset, bt))
                        {
                            rp.offset.x += e.offsetX;
                            rp.offset.y += e.offsetY;
                            rp.scaling.x *= e.scaleX;
                            rp.scaling.y *= e.scaleY;
                        }
                    }
                }
            }
        }

        public override IEnumerable<string> ConfigErrors()
        {
            Reload();
            yield break;
        }
    }

    public class QOLDefModExtension_RenderingParametersConfig
    {
        public List<QOLDefModExtension_RenderingParametersConfig_Entry> entries = new List<QOLDefModExtension_RenderingParametersConfig_Entry>();
    }

    public class QOLDefModExtension_RenderingParametersConfig_Entry
    {
        public BodyTypeDef bodyType = null;
        public int facing = 4;
        public float offsetX = 0f, offsetY = 0f, scaleX = 1f, scaleY = 1f;
        public bool matches(int f, BodyTypeDef bt)
        {
            return (f == facing || facing == 4) && (bodyType == null || bodyType == bt);
        }
    }
}
