﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 10.10.2018
 * Time: 14:55
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using RIMMSqol.remnantcolony.listeners;
using RIMMSqol.remnantcolony.pawnfilters;
using RIMMSqol.remnantcolony.pawnvalues;
using RIMMSqol.translations;
using RimWorld;
using Verse;

namespace RIMMSqol.remnantcolony
{
	/// <summary>
	/// Description of QOLCityDef.
	/// </summary>
	public class QOLCityDef : QOLCityListenerOwnerDef {
		public List<QOLCityProjectDef> projects;
		public List<QOLCityNeedDef> needs;
		public List<QOLSocietyIndicatorDef> indicators;
		public List<QOLCityResourceDef> resources;
		public QOLCityPopulationProperties population;
		public QOLCityNeedDef labourDef;
		public TraderKindDef traderKind;
		public WorldObjectDef worldObject;
		
		public override void PostLoad() {
			base.PostLoad();
			if ( indicators == null ) indicators = new List<QOLSocietyIndicatorDef>();
			if ( needs == null ) needs = new List<QOLCityNeedDef>();
			if ( resources == null ) resources = new List<QOLCityResourceDef>();
			if ( projects == null ) projects = new List<QOLCityProjectDef>();
			if ( population == null ) population = new QOLCityPopulationProperties();
		}
		
		public override IEnumerable<string> ConfigErrors() {
			if ( labourDef == null ) {
				labourDef = DefDatabase<QOLCityNeedDef>.GetNamed("QOLLabour", false);
				if ( labourDef == null ) {
					labourDef = DefDatabase<QOLCityNeedDef>.AllDefsListForReading[0];
				}
			}
			if ( traderKind == null ) {
				traderKind = DefDatabase<TraderKindDef>.GetNamed("RemnantColony", false);
				if ( traderKind == null ) {
					Log.Error("QOLCity def \""+defName+"\" is missing trader kind definition and default was not found!");
					traderKind = DefDatabase<TraderKindDef>.AllDefsListForReading[0];
				}
			}
			return base.ConfigErrors();
		}
	}
	
	public class QOLCityPopulationProperties {
		public List<QOLCitySocialStandingDef> socialStandings;
		public List<QOLCityPawnFilter> immigrationLaws;
		public List<QOLCityPopulationContribution> contributions;
		public List<QOLCityPawnProperties> properties;
		public bool IsImmigrationAllowed(QOLCity city, Pawn p) {
			return !immigrationLaws.Any(il=>!il.IsAllowed(city,p));
		}
		public void CityContribution(QOLCity city, QOLCityState state) {
			foreach ( QOLCityPopulationContribution contribution in contributions ) {
				contribution.CityContribution(city,state);
			}
		}
		public Dictionary<string,QOLCityValueMutator> calculateProperties(QOLCityState state, Pawn p) {
			Dictionary<string,QOLCityValueMutator> dic = new Dictionary<string, QOLCityValueMutator>(properties.Count);
			foreach ( QOLCityPawnProperties property in properties ) {
				if ( !property.key.NullOrEmpty() ) {
					dic.Add(property.key,property.getMutator(state,p));
				}
			}
			return dic;
		}
	}
	
	public class QOLCityPawnProperties {
		public string key;
		public QOLCityPawnValue flat;
		public QOLCityPawnValue factor;
		public float constantFlat = 0f;
		public float constantFactor = 1f;
		public QOLCityValueMutator getMutator(QOLCityState state, Pawn p) {
			return QOLCityValueMutator.create(flat != null ? flat.Value(state,p) : constantFlat, 
			                                  factor != null ? factor.Value(state,p) : constantFactor);
		}
	}
	
	public class QOLCityPopulationContribution : QOLCityStateContribution {
		public List<QOLCityPawnFilter> pawnFilters;
		public List<QOLCityPawnValue> pawnValues;
		public float multiplierPopulationCount;
		protected float RelevantPopulationCount(QOLCity city, QOLCityState state) {
			float relevantPopulationCount = 0;
			if ( !pawnFilters.NullOrEmpty() ) {
				foreach ( Pawn p in city.population.citizens.Keys ) {
					if ( pawnFilters.All(filter => filter.IsAllowed(city, p)) ) {
						relevantPopulationCount += pawnValues.NullOrEmpty() ? 1 : pawnValues.Sum(pv => pv.Value(state, p));
					}
				}
			} else {
				relevantPopulationCount = city.population.citizens.Count;
			}
			return relevantPopulationCount * multiplierPopulationCount;
		}
		public void CityContribution(QOLCity city, QOLCityState state) {
			QOLCityStateContribution adjustedState = new QOLCityStateContribution();
			
			float multiplier = RelevantPopulationCount(city, state);
			foreach ( QOLCityStateContributionNeed need in needs ) {
				adjustedState.needs.Add(new QOLCityStateContributionNeed() {
					def = need.def, 
					required = new NumericAdjustment(){flat = need.required.flat * multiplier, factor = need.required.factor * multiplier}, 
					provided = new NumericAdjustment(){flat = need.provided.flat * multiplier, factor = need.provided.factor * multiplier}});
			}
			foreach ( QOLCityStateContributionIndicator indicator in indicators ) {
				List<QOLCityPawnFilter> joinedPawnFilters;
				if ( pawnFilters.NullOrEmpty() ) {
					if ( !indicator.pawnFilters.NullOrEmpty() )
						joinedPawnFilters = new List<QOLCityPawnFilter>(indicator.pawnFilters);
					else joinedPawnFilters = null;
				} else {
					joinedPawnFilters = new List<QOLCityPawnFilter>(pawnFilters);
					if ( !indicator.pawnFilters.NullOrEmpty() )
						joinedPawnFilters.AddRange(indicator.pawnFilters);
				}
				adjustedState.indicators.Add(new QOLCityStateContributionIndicator(indicator.def, indicator.Stage, joinedPawnFilters, indicator.affinity*multiplierPopulationCount, indicator.suppression*multiplierPopulationCount));
			}
			foreach ( QOLCityStateContributionResource resource in resources ) {
				adjustedState.resources.Add(new QOLCityStateContributionResource() {
					def = resource.def, 
					flat = resource.flat * multiplier,
					factor = resource.factor * multiplier});
			}
			foreach ( QOLCityStateContributionThing thing in things ) {
				adjustedState.things.Add(new QOLCityStateContributionThing() {
					def = thing.def, 
					flat = thing.flat * multiplier,
					factor = thing.factor * multiplier});
			}
			state.Add(adjustedState);
		}
	}
}
