﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 08.10.2018
 * Time: 15:55
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Text;
using RIMMSqol.remnantcolony.actions;
using RIMMSqol.remnantcolony.listeners;
using RIMMSqol.remnantcolony.pawnfilters;
using RIMMSqol.remnantcolony.requirements;
using RIMMSqol.translations;
using RimWorld;
using Verse;

namespace RIMMSqol.remnantcolony
{
	/// <summary>
	/// Description of QOLCityProject.
	/// </summary>
	public class QOLCityProjectDef : QOLCityListenerNonOwnerDef
	{
		[DefaultValue(null), Description("World object def that is used as a template to construct a world object representing a completed project in a city.")]
		public WorldObjectDef worldObject;
		[DefaultValue(int.MaxValue), Description("Used to sort this def in relation to other instances of this def. Default order is from low to high.")]
		public int order = int.MaxValue;
		[DefaultValue(0), Description("A positive number above zero indicates that this project last for that number of turns before becoming inactive.")]
		public int lifetime = 0;
		[DefaultValue(0), Description("A positive number indicates that this project is started after that number of turns. A negative number means it is started periodically after that number of turns until maxInstances are reached. Zero means it is build manually.")]
		public int autospawnTimer = 0;
		[DefaultValue(false), Description("Whether a currently being built or already existing project can be removed by the user.")]
		public bool cancellable = false;
		[DefaultValue(0), Description("How much labour is required to finish this project.")]
		public float labour = 0;
		[DefaultValue(1), Description("How often can this project be repeated in one city.")]
		public int maximumInstances = 1;
		[DefaultValue(null), Description("Conditions that must be met before the project can be started.")]
		public List<QOLCityRequirement> requirements;
		[DefaultValue(false), Description("Whether the project can be activated/deactivated.")]
		public bool toggleable = false;
		[DefaultValue(null), Description("Describes the influence the project has in its various states.")]
		public QOLCityStateContributions states;
		[DefaultValue(null), Description("Lists all actions provided by this project.")]
		public List<QOLCityActionDef> actions;
		[DefaultValue(null), Description("Can be provided to use a custom class when creating the state object for this configuration. If its null the default object will be used.")]
		public Type stateClass;
		[DefaultValue(null), Description("How and where to display the project.")]
		public QOLCityProjectCategoryDef category;
		[DefaultValue(true), Description("Whether a notification is sent when the project is built.")]
		public bool sendCompletedNotification = true;
		[DefaultValue(false), Description("Whether a project should be set to inactive after it was built.")]
		public bool spawnsInactive = false;
		/*[DefaultValue(null), Description("Lists all actions accessible when examining an instance of this project.")]
		public List<QOLCityActionDef> detailActions;*/
		
		public override void PostLoad() {
			base.PostLoad();
			if ( requirements == null ) requirements = new List<QOLCityRequirement>(0);
			if ( states == null ) states = new QOLCityStateContributions();
			if ( actions == null ) actions = new List<QOLCityActionDef>();
			//if ( detailActions == null ) detailActions = new List<QOLCityActionDef>();
		}
		
		public bool AllowsNumberOfInstances(int noInstances) {
			if ( maximumInstances <= 0 ) return true;
			return noInstances <= maximumInstances;
		}
		
		public virtual string Tooltip () {
			StringBuilder sb = new StringBuilder();
			
			sb.AppendLine(label);
			if ( !requirements.NullOrEmpty() ) {
				sb.AppendLine("Requirements");
				foreach ( QOLCityRequirement requirement in requirements ) {
					sb.AppendLine(requirement.Description());
				}
			}
			if ( !states.build.indicators.NullOrEmpty() || !states.build.resources.NullOrEmpty() || !states.build.needs.NullOrEmpty() || !states.build.things.NullOrEmpty() ) {
				sb.AppendLine("Phase: Building");
				states.build.ToolTip(sb);
			}
			if ( !states.active.indicators.NullOrEmpty() || !states.active.resources.NullOrEmpty() || !states.active.needs.NullOrEmpty() || !states.active.things.NullOrEmpty() ) {
				sb.AppendLine("Phase: Active");
				states.active.ToolTip(sb);
			}
			if ( !states.inactive.indicators.NullOrEmpty() || !states.inactive.resources.NullOrEmpty() || !states.inactive.needs.NullOrEmpty() || !states.inactive.things.NullOrEmpty() ) {
				sb.AppendLine("Phase: Inactive");
				states.inactive.ToolTip(sb);
			}
			sb.AppendLine(description);
			if ( cancellable ) sb.AppendLine("Can be manually removed");
			else sb.AppendLine("Can not be manually removed");
			if ( toggleable ) sb.AppendLine("Can be deactivated");
			else sb.AppendLine("Can not be deactivated");
			if ( lifetime > 0 ) sb.AppendLine("Times out after "+lifetime+" turns.");
			
			if ( autospawnTimer > 0 ) sb.AppendLine("Spawns after "+autospawnTimer+" turns.");
			if ( autospawnTimer < 0 ) sb.AppendLine("Spawns every "+autospawnTimer+" turns.");
			if ( autospawnTimer == 0 ) sb.AppendLine("Is manually built.");
			
			if ( labour > 0 ) sb.AppendLine("Work needed to built project "+labour);
			if ( maximumInstances > 0 ) sb.AppendLine("Limited to "+maximumInstances+" instances.");
			if ( worldObject != null ) sb.AppendLine("Requires an empty world tile.");
			
			return sb.ToString();
		}
	}
	
	public class QOLCityStateContributions {
		public QOLCityStateContribution build;
		public QOLCityStateContribution active;
		public QOLCityStateContribution inactive;
		public QOLCityStateContributions() {
			build = new QOLCityStateContribution();
			active = new QOLCityStateContribution();
			inactive = new QOLCityStateContribution();
		}
	}
	
	public static class QOLCityStateContributionExtension {
		static public bool NullOrEmpty(this QOLCityStateContribution sc) {
			return sc == null || (sc.needs.NullOrEmpty() && sc.indicators.NullOrEmpty() && sc.resources.NullOrEmpty() && sc.things.NullOrEmpty());
		}
	}
	
	public class QOLCityStateContribution {
		public List<QOLCityStateContributionIndicator> indicators;
		public List<QOLCityStateContributionNeed> needs;
		public List<QOLCityStateContributionResource> resources;
		public List<QOLCityStateContributionThing> things;
		public QOLCityStateContribution() {
			indicators = new List<QOLCityStateContributionIndicator>();
			needs = new List<QOLCityStateContributionNeed>();
			resources = new List<QOLCityStateContributionResource>();
			things = new List<QOLCityStateContributionThing>();
		}
		
		public void ToolTip(StringBuilder sb) {
			foreach ( QOLCityStateContributionThing thing in things ) {
				if ( Math.Abs(thing.flat) > float.Epsilon ) {
					sb.Append(thing.flat).Append(" ").AppendLine(thing.def.label);
				}
				if ( Math.Abs(thing.factor) > float.Epsilon ) {
					sb.Append(thing.factor.ToString("P0")).Append(" ").AppendLine(thing.def.label);
				}
			}
			foreach ( QOLCityStateContributionResource resource in resources ) {
				if ( Math.Abs(resource.flat) > float.Epsilon ) {
					sb.Append(resource.flat).Append(" ").AppendLine(resource.def.label);
				}
				if ( Math.Abs(resource.factor) > float.Epsilon ) {
					sb.Append(resource.factor.ToString("P0")).Append(" ").AppendLine(resource.def.label);
				}
			}
			foreach ( QOLCityStateContributionNeed need in needs ) {
				if ( Math.Abs(need.required.flat) > float.Epsilon ) {
					sb.Append("Requires ").Append(need.required.flat).Append(" ").AppendLine(need.def.label);
				}
				if ( Math.Abs(need.required.factor) > float.Epsilon ) {
					sb.Append("Requires ").Append(need.required.factor.ToString("P0")).Append(" ").AppendLine(need.def.label);
				}
				if ( Math.Abs(need.provided.flat) > float.Epsilon ) {
					sb.Append("Provides ").Append(need.provided.flat).Append(" ").AppendLine(need.def.label);
				}
				if ( Math.Abs(need.provided.factor) > float.Epsilon ) {
					sb.Append("Provides ").Append(need.provided.factor.ToString("P0")).Append(" ").AppendLine(need.def.label);
				}
			}
			foreach ( QOLCityStateContributionIndicator indicator in indicators ) {
				sb.Append(indicator.def.label).Append("/").Append(indicator.Stage.label).AppendLine(":");
				if ( !indicator.pawnFilters.NullOrEmpty() ) {
					sb.AppendLine("Requirements for citizens");
					foreach ( QOLCityPawnFilter filter in indicator.pawnFilters ) {
						sb.AppendLine(filter.Description());
					}
				}
				if ( Math.Abs(indicator.affinity) > float.Epsilon ) {
					sb.Append("+").AppendLine(indicator.affinity.ToString("P0"));
				}
				if (Math.Abs(indicator.suppression) > float.Epsilon ) {
					sb.Append("-").Append(indicator.suppression.ToString("P0"));
				}
			}
		}
	}
	
	public class QOLCityStateContributionResource : NumericAdjustment {
		public QOLCityResourceDef def;
		override public void ExposeData() {
			base.ExposeData();
			Scribe_Defs.Look<QOLCityResourceDef>(ref def,"def");
		}
	}
	
	public class QOLCityStateContributionThing : NumericAdjustment {
		public ThingDef def;
        public QualityCategory quality = QualityCategory.Normal;
		override public void ExposeData() {
			base.ExposeData();
			Scribe_Defs.Look<ThingDef>(ref def,"def");
            Scribe_Values.Look<QualityCategory>(ref quality, "quality", QualityCategory.Normal);
        }
	}
	
	public class QOLCityStateContributionIndicator {
		public QOLSocietyIndicatorDef def;
		public List<QOLCityPawnFilter> pawnFilters;
		protected string stageId = null;
		public float affinity = 0, suppression = 0;
		[Unsaved]
		protected QOLSocietyIndicatorStage stageCached;
		public QOLSocietyIndicatorStage Stage {
			get {
				if ( stageCached == null && def != null && !def.stages.NullOrEmpty() ) {
					stageCached = def.stages.FirstOrFallback<QOLSocietyIndicatorStage>(s=>s.id == stageId, null);
					if ( stageCached == null ) {
						stageCached = def.stages[0];
					}
				}
				return stageCached;
			}
		}
		
		public QOLCityStateContributionIndicator() {}
		public QOLCityStateContributionIndicator(QOLSocietyIndicatorDef def, QOLSocietyIndicatorStage stage, List<QOLCityPawnFilter> pawnFilters, 
		                                         float affinity, float suppression) {
			this.def = def;
			this.stageId = stage.id;
			this.stageCached = stage;
			this.affinity = affinity;
			this.suppression = suppression;
			this.pawnFilters = pawnFilters;
		}
	}
	
	public class QOLCityStateContributionNeed : ProvidedAndRequired {
		public QOLCityNeedDef def;
		override public void ExposeData() {
			base.ExposeData();
			Scribe_Defs.Look<QOLCityNeedDef>(ref def,"def");
		}
	}
}
