﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 13.12.2018
 * Time: 19:21
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections;
using System.Linq;
using System.Collections.Generic;
using RIMMSqol.renderers;
using RIMMSqol.extensions;
using RimWorld;
using UnityEngine;
using Verse;

namespace RIMMSqol.remnantcolony.actions
{
	/// <summary>
	/// Description of QOLCityAction_Craft.
	/// </summary>
	public class QOLCityAction_Craft : QOLCityAction {
		protected int workAllocatedInt;
		public QOLCityActionConfiguration_Craft conf;
		public int WorkAllocated { get {return workAllocatedInt;} }
		public List<CraftOrder> orders;
		
		public QOLCityAction_Craft() : base(null) {}
		
		public QOLCityAction_Craft(QOLCityActionDef def) : base(def) {
			workAllocatedInt = 0;
			conf = def.action as QOLCityActionConfiguration_Craft;
			if ( conf == null ) {
				Log.Error("No matching conf found in "+typeof(QOLCityAction_Craft).Name+"!");
			}
		}

		public override void ExposeData() {
			base.ExposeData();
			Scribe_Values.Look<int>(ref workAllocatedInt, "workAllocated", 0);
			Scribe_Collections.Look<CraftOrder>(ref orders, "orders", LookMode.Deep, null);
			if (Scribe.mode == LoadSaveMode.ResolvingCrossRefs) {
				if ( def != null ) {
					conf = def.action as QOLCityActionConfiguration_Craft;
					if ( conf == null ) {
						Log.Error("No matching conf found in "+typeof(QOLCityAction_Craft).Name+"!");
					}
				} else {
					Log.Error("No definition found in "+typeof(QOLCityAction_Craft).Name+"!");
				}
			}
		}
		
		public override void OnClick(QOLCity city, Pawn negotiator) {
			Find.WindowStack.Add(new QOLCityActionDialog_Craft(this,city));
		}
		
		public void IncreaseWork() {
			workAllocatedInt = Math.Min(Math.Max(0,++workAllocatedInt),instances*conf.maxLabourPerInstance);
		}
		
		public void DecreaseWork() {
			workAllocatedInt = Math.Max(0,--workAllocatedInt);
		}
	}
	
	public class QOLCityActionConfiguration_Craft : QOLCityActionConfiguration {
		public float workMultiplier = 1f;
		public int maxLabourPerInstance = 1;
		public float factorWorkAwful = 0.5f,
			factorWorkPoor = 0.7f,
			factorWorkNormal = 1f,
			factorWorkGood = 1.2f,
			factorWorkExcellent = 1.4f,
			factorWorkMasterwork = 1.6f,
			factorWorkLegendary = 2f;
		public List<ThingCategoryDef> categories;
		public List<QualityCategory> allowedQualityCategories = new List<QualityCategory>();
		
		public List<ThingCategoryDef> AllCategories {
			get {
				if ( categories != null ) return (from cs in categories from c in cs.ThisAndChildCategoryDefs select c).Distinct().ToList();
				return new List<ThingCategoryDef>();
			}
		}
		
		public override QOLCityAction CreateAction(QOLCityActionDef def) {
			return new QOLCityAction_Craft(def);
		}
		
		public float QualityWorkMultiplier(QualityCategory quality) {
			switch (quality) {
				case QualityCategory.Awful:
					return factorWorkAwful;
				case QualityCategory.Poor:
					return factorWorkPoor;
				case QualityCategory.Normal:
					return factorWorkNormal;
				case QualityCategory.Good:
					return factorWorkGood;
				case QualityCategory.Excellent:
					return factorWorkExcellent;
				case QualityCategory.Masterwork:
					return factorWorkMasterwork;
				case QualityCategory.Legendary:
					return factorWorkLegendary;
				default:
					throw new ArgumentOutOfRangeException();
			}
		}
	}
	
	public class CraftOrder : IExposable {
		protected RecipeDef _recipe = null;
		public QualityCategory quality = QualityCategory.Normal;
		public int amount = 1;
		public int workLeft;
		public List<Pair<ThingDef, IngredientCount>> ingredients = new List<Pair<ThingDef, IngredientCount>>();
		public List<ThingDefCountClass> products = new List<ThingDefCountClass>();
		
		public RecipeDef recipe {
			get { return _recipe; }
			set {
				if ( value == null ) {
					if ( _recipe != null ) {
						_recipe = value;
						ingredients.Clear();
						products.Clear();
					}
				} else if ( !value.Equals(_recipe) ) {
					_recipe = value;
					ingredients.Clear();
					products.Clear();
					foreach ( IngredientCount ic in recipe.ingredients ) {
						ingredients.Add(new Pair<ThingDef, IngredientCount>(ic.filter.AllowedThingDefs.First(),ic));
					}
					foreach ( ThingDefCountClass product in recipe.products ) {
						products.Add(new ThingDefCountClass(product.thingDef,product.count));
					}
				}
			}
		}

		public void ExposeData() {
			Scribe_Defs.Look<RecipeDef>(ref _recipe, "recipe");
			Scribe_Values.Look<int>(ref workLeft, "workLeft", 0);
			Scribe_Values.Look<int>(ref amount, "amount", 0);
			Scribe_Values.Look<QualityCategory>(ref quality, "quality", QualityCategory.Normal);
			Scribe_Collections.Look<ThingDefCountClass>(ref products, "products", LookMode.Deep, null);
			Scribe_Collections.Look<Pair<ThingDef, IngredientCount>>(ref ingredients, "ingredients", LookMode.Deep, null);
		}
		
		public bool HasQuality() {
			return products.Any(p=>p.thingDef.HasComp(typeof(CompQuality)));
		}
		public bool HasStuff() {
			return products.Any(p=>p.thingDef.MadeFromStuff);
		}
		public int stockCount(IThingHolder th, ThingDef thing) {
			try {
				return th.GetDirectlyHeldThings().Where(t => t.def.defName == thing.defName).Sum(t => t.stackCount);
			} catch {
				return 0;
			}
		}
		public bool canAfford(IThingHolder th) {
			if ( recipe == null ) return false;
			foreach ( Pair<ThingDef,IngredientCount> p in ingredients ) {
				if ( stockCount(th, p.First) < p.Second.CountRequiredOfFor(p.First,recipe) * amount) {
					return false;
				}
			}
			return true;
		}
		public bool pay(QOLCity city, QOLCityAction_Craft action) {
			if ( canAfford(city) ) {
				bool paidInFull = true;
				ThingOwner owner = city.GetDirectlyHeldThings();
				
				foreach ( Pair<ThingDef,IngredientCount> p in ingredients ) {
					int countToGive = p.Second.CountRequiredOfFor(p.First,recipe) * amount;
					Thing thingToSplit = null;
					owner.RemoveAll(t=>{ 
					                	if ( p.First==t.def ) {
					                		if ( countToGive >= t.stackCount ) {
					                			countToGive -= t.stackCount; 
					    						return true;
					                		}
					                		thingToSplit = t;
						                	}
						                	return false;
						                });
					if ( countToGive > 0 ) {
						if ( thingToSplit != null ) {
							Thing thingRemoved = owner.Take(thingToSplit,countToGive);
							if ( thingRemoved == null || thingRemoved.stackCount != countToGive ) {
								paidInFull = false;
								break;
							}
						}
					}
				}
				if ( !paidInFull ) return false;
				
				ThingDef stuff = ingredients.First(p=>p.First.IsStuff).First;
				workLeft = Mathf.CeilToInt(recipe.WorkAmountTotal(stuff) * amount * action.conf.QualityWorkMultiplier(quality));
				return true;
			} else {
				return false;
			}
		}
		public void giveProductsTo(ITrader trader) {
			foreach ( Thing t in createProducts() ) {
				trader.GiveSoldThingToTrader(t,t.stackCount,null);
			}
		}
		protected List<Thing> createProducts() {
			List<Thing> generatedProducts = new List<Thing>();
			for ( int i = 0; i < amount; i++ ) {
				foreach ( ThingDefCountClass tdcc in products ) {
					Thing t;
					if ( tdcc.thingDef.MadeFromStuff ) {
						t = ThingMaker.MakeThing(tdcc.thingDef, ingredients.First(p=>p.First.IsStuff).First);
					} else {
						t = ThingMaker.MakeThing(tdcc.thingDef, null);
					}
					t.stackCount = tdcc.count;
					MinifiedThing minifiedThing = t as MinifiedThing;
					CompQuality compQuality = (minifiedThing == null) ? t.TryGetComp<CompQuality>() : minifiedThing.InnerThing.TryGetComp<CompQuality>();
					if ( compQuality != null ) compQuality.SetQuality(quality, ArtGenerationContext.Outsider);
					generatedProducts.Add(t);
				}
			}
			return generatedProducts;
		}
	}
	
	public class QOLCityActionDialog_Craft : Dialog_Flow {
		protected readonly Color? titleColor = new Color(0.2f, 0.8f, 0.85f);
		
		protected QOLCity city;
		protected QOLCityAction_Craft action;
		protected List<QualityCategory> allowedQualityCategories = new List<QualityCategory>();
		
		protected string titleSelectOrder = "QOLCityActionCraftSelectOrder";
		protected string titleEditOrder = "QOLCityActionCraftEditOrder";
		protected string titleAssignWork = "QOLCityActionRepairAllocateLabour";
		protected string labelRecipe = "QOLCityActionCraftRecipe";
		protected string labelChoose = "WorldChooseButton";
		protected string labelQuality = "Quality";
		protected string labelStuff = "StatsReport_Material";
		protected string labelNone = "None";
		protected string labelIngredients = "Ingredients";
		protected string labelProducts = "Products";
		protected string labelPlaceOrder = "AcceptButton";
		protected string labelAmount = "Amount";
		
		public QOLCityActionDialog_Craft(QOLCityAction_Craft action, QOLCity city) : base(null,400f+18f*2,QOLMod.LineHeight(GameFont.Medium)*20+18f*2) {
			this.action = action;
			this.allowedQualityCategories = action.conf.allowedQualityCategories;
			this.city = city;
			
			titleSelectOrder.TryTranslate(out titleSelectOrder);
			titleEditOrder.TryTranslate(out titleEditOrder);
			titleAssignWork.TryTranslate(out titleAssignWork);
			labelRecipe.TryTranslate(out labelRecipe);
			labelChoose.TryTranslate(out labelChoose);
			labelQuality.TryTranslate(out labelQuality);
			labelStuff.TryTranslate(out labelStuff);
			labelNone.TryTranslate(out labelNone);
			labelIngredients.TryTranslate(out labelIngredients);
			labelProducts.TryTranslate(out labelProducts);
			labelPlaceOrder.TryTranslate(out labelPlaceOrder);
			
			Dictionary<string,PageRenderer> pages = new Dictionary<string, PageRenderer>();
			pages.Add("menu", new PageRenderer(f=>"", null, null)
			          .AddChild(new RowLayoutRenderer()
			                    .AddChild(new ButtonTextRenderer(titleAssignWork, f=>f.navigate("assignWork"), GameFont.Medium))
			                    .AddChild(new ButtonTextRenderer(titleSelectOrder, f=>f.navigate("selectOrder"), GameFont.Medium))));
			
			pages.Add("assignWork", new PageRenderer(f=>titleAssignWork, null, ()=>"menu")
			          .AddChild(new HorizontalSplitRenderer(
			          	new ColumnLayoutRenderer()
			                    .AddChild(new ButtonTextRenderer("-", f => this.action.DecreaseWork(), GameFont.Medium, 32f))
			                    .AddChild(new LabelRenderer(f=>this.action.WorkAllocated.ToString("###0"), 322f, GameFont.Medium, TextAnchor.MiddleCenter))
			                    .AddChild(new ButtonTextRenderer("+", f => this.action.IncreaseWork(), GameFont.Medium, 32f)),
	                    new EmptyRenderer(),-1f,32f)));
			
			pages.Add("selectOrder", new PageRenderer(f=>titleSelectOrder, null, ()=>"menu", null)
			          .AddChild(new RowLayoutRenderer()
			                    .AddChild(new ButtonTextRenderer("New Order",f=>{
			                                           				f.flowScope["selectItem"] = new CraftOrder();
			                                                     	f.navigate("editOrder");
			                                                     }, GameFont.Medium))
			                    .AddChild(new ListRenderer(f=>action.orders,-1,200f)
			                              .AddChild(new ColumnLayoutRenderer()
			                                        .AddChild(new ButtonTextRenderer(f=>{
			                                                     	CraftOrder order = ((IterationItem)f.pageScope["curItem"]).curItem as CraftOrder;
			                                                     	if ( order != null ) {
			                                                     		if ( order.recipe != null ) 
			                                                     			return order.amount+" X "+order.recipe.label+" ("+order.workLeft+")";
			                                                     		return titleSelectOrder;
			                                                     	}
			                                                     	return titleEditOrder;
			                                                     }, f=>{
			                                                     	f.flowScope["selectItem"] = ((IterationItem)f.pageScope["curItem"]).curItem;
			                                                     	f.navigate("viewOrder");
			                                                     }, GameFont.Medium, 328f))
			                              			.AddChild(new ButtonTextRenderer("+",f=>{
			                                                               	CraftOrder order = ((IterationItem)f.pageScope["curItem"]).curItem as CraftOrder;
			                                                     			if ( order != null ) {
			                                                               		f.addPostRenderCallback(flow=>{
			                                                               		                        	int indx = action.orders.IndexOf(order);
			                                                               		                        	if ( indx > 0 ) {
			                                                               		                        		if ( action.orders.Remove(order) ) {
			                                                               		                        			action.orders.Insert(indx-1,order);
			                                                               		                        		}
			                                                               		                        	}
			                                                               		                        });
			                                                               	}
			                                                               }, GameFont.Medium, 32f))
			                              			.AddChild(new ButtonTextRenderer("-",f=>{
			                                                               	CraftOrder order = ((IterationItem)f.pageScope["curItem"]).curItem as CraftOrder;
			                                                     			if ( order != null ) {
			                                                               		f.addPostRenderCallback(flow=>{
			                                                               		                        	int indx = action.orders.IndexOf(order);
			                                                               		                        	if ( indx < action.orders.Count-1 ) {
			                                                               		                        		if ( action.orders.Remove(order) ) {
			                                                               		                        			action.orders.Insert(indx+1,order);
			                                                               		                        		}
			                                                               		                        	}
			                                                               		                        });
			                                                               	}
                                                                         }, GameFont.Medium,32f))))));

			pages.Add("editOrder", new PageRenderer(f=>titleSelectOrder, null, ()=>"selectOrder", null)
			          .AddChild(new VerticalSplitRenderer(new RowLayoutRenderer()
                              	.AddChild(new StyleRenderer(new LabelRenderer(labelRecipe,-1f,GameFont.Medium,TextAnchor.MiddleCenter,titleColor)))
			                    .AddChild(new ButtonTextRenderer(f=>{
			                                                     	CraftOrder order = f.flowScope["selectItem"] as CraftOrder; 
			                                                     	if ( order != null ) {
			                                                     		return order.recipe == null ? labelChoose : order.recipe.label;
			                                                     	}
			                                                     	return "error";
			                                                     },
			                                                     f=>{
			                                                     	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                     	if ( order != null ) {
				                                                     	Find.WindowStack.Add(
			                                                     			new Dialog_Select<RecipeDef>(def=>{f.addPostRenderCallback(ff=>order.recipe = def); return true;},
						                                                     	                             DefDatabase<RecipeDef>.AllDefsListForReading
						                                                     	                             .Where(d=>d.AvailableNow && !d.products.NullOrEmpty() &&
						                                                     	                                    d.products.
						                                                     	                                    All(tdcc=>tdcc.thingDef.thingCategories != null && 
						                                                     	                                        action.conf.AllCategories.
						                                                     	                                        SharesElementWith(tdcc.thingDef.thingCategories)))
						                                                     	                             	.OrderBy(d=>d.label),
						                                                     	                             d=>d.label)
				                                                     	);
			                                                     	}
			                                                     }, GameFont.Medium))
			                    .AddChild(new LabelRenderer(labelQuality,-1f,GameFont.Medium,TextAnchor.MiddleCenter,titleColor))
			                    .AddChild(new HideRenderer(f=>{
			                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                               	return order == null || !order.HasQuality();
			                                               },
			                                               new ButtonTextRenderer(f => (f.flowScope["selectItem"] as CraftOrder).quality.GetLabel(),
			                                                                      f=>{
			                                                                      	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                                       	Find.WindowStack.Add(new Dialog_Select<QualityCategory>(q=>{
			                                                                       	                                                        	order.quality = q; 
			                                                                       	                                                        	return true;
			                                                                      	                                                        }, allowedQualityCategories, d=>d.GetLabel()));
			                                                                      }, GameFont.Medium)))
			                    .AddChild(new LabelRenderer(labelStuff,-1f,GameFont.Medium,TextAnchor.MiddleCenter,titleColor))
			                    .AddChild(new LabelRenderer(f=>{
			                                                	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                	if ( order != null ) {
			                                                		if ( order.HasStuff() )
			                                                			return order.ingredients.First(p=>p.First.IsStuff).First.label;
			                                                		return labelNone;
			                                                	}
			                                                	return "error";
			                                                },-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    
			                    .AddChild(new LabelRenderer(labelAmount,-1f,GameFont.Medium,TextAnchor.MiddleCenter,titleColor))
			                    .AddChild(new ColumnLayoutRenderer()
			                              .AddChild(new ButtonTextRenderer("-", f=>{
			                                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                               	if ( order != null ) order.amount = Math.Max(1,--order.amount);
			                                                               }, GameFont.Medium, 32f))
			                              .AddChild(new LabelRenderer(f=>{
			                                                          	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                          	if ( order != null ) {
			                                                          		return order.amount.ToString("0");
			                                                          	}
			                                                          	return "error";
			                                                          },128f,GameFont.Medium,TextAnchor.MiddleCenter))
			                              .AddChild(new ButtonTextRenderer("+", f=>{
			                                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                               	if ( order != null ) order.amount++;
			                                                               }, GameFont.Medium, 32f))
			                             )
			                    .AddChild(new ButtonTextRenderer(labelPlaceOrder, f=>{
			                                                     	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                     	if ( order != null ) {
			                                                     		if ( order.pay(city,action) ) {
			                                                     			f.addPostRenderCallback(ff=> {
			                                                     			                        	if ( action.orders != null ) action.orders.Add(order);
			                                                     			                        	else {
			                                                     			                        		action.orders = new List<CraftOrder>();
			                                                     			                        		action.orders.Add(order);
			                                                     			                        	}
			                                                     			                        });
			                                                     			f.navigate("selectOrder");
			                                                     		}
			                                                     	}
			                                                     }, GameFont.Medium, 32f))
			                    ,new RowLayoutRenderer()
			                    .AddChild(new LabelRenderer(labelIngredients,-1f,GameFont.Medium,TextAnchor.MiddleCenter,titleColor))
			                    .AddChild(new ListRenderer(f=>{
			                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                               	if ( order != null ) return order.ingredients;
			                                               	return new List<Pair<ThingDef, IngredientCount>>();
			                                               }, -1f, QOLMod.LineHeight(GameFont.Medium)*3)
			                              .AddChild(new ButtonTextRenderer(f => {
                                                                        	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                               	Pair<ThingDef, IngredientCount> p = ((Pair<ThingDef, IngredientCount>)((IterationItem)f.pageScope["curItem"]).curItem);
			                                                               	ThingDef tDef = p.First;
			                                                               	int crf = p.Second.CountRequiredOfFor(p.First, order.recipe);
			                                                               	int sc = order.stockCount(city,tDef);
			                                                               	return tDef.label+" "+crf.ToString()+"("+sc.ToString()+")";
			                                                               },
			                                                               f=>{
			                                                               	f.pageScope["selectItem"] = ((IterationItem)f.pageScope["curItem"]).curItem; 
			                                                               	Find.WindowStack.Add(new Dialog_SelectDef<ThingDef>(def=>{
			                                                               	                                                    	Pair<ThingDef,IngredientCount> oldPair = ((Pair<ThingDef,IngredientCount>)f.pageScope["selectItem"]);
			                                                               	                                                    	Pair<ThingDef,IngredientCount> newPair = new Pair<ThingDef, IngredientCount>(def,oldPair.Second);
			                                                               	                                                    	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                               	                                                    	int indx = order.ingredients.IndexOf(oldPair);
			                                                               	                                                    	order.ingredients.Remove(oldPair);
			                                                               	                                                    	order.ingredients.Insert(indx,newPair);
			                                                               	                                                    	
			                                                               	                                                    	f.pageScope["selectItem"] = null; 
			                                                               	                                                    	return true;
			                                                               	                                                    },
			                                                               	                                                    ((Pair<ThingDef,IngredientCount>)f.pageScope["selectItem"]).Second.filter.AllowedThingDefs));
			                                                               }, GameFont.Medium, 200f)))
			                    
			                    .AddChild(new LabelRenderer(labelProducts,-1f,GameFont.Medium,TextAnchor.MiddleCenter,titleColor))
			                    .AddChild(new ListRenderer(f=>{
			                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                               	if ( order != null ) return order.products;
			                                               	return new List<ThingDefCountClass>();
			                                               }, -1, QOLMod.LineHeight(GameFont.Medium)*3)
			                                               .AddChild(new LabelRenderer(f=>{
			                                                                           	ThingDefCountClass tdcc = ((ThingDefCountClass)((IterationItem)f.pageScope["curItem"]).curItem);
			                                                                           	if ( tdcc != null ) return tdcc.thingDef.label+" "+tdcc.count;
			                                                                           	return "error";
			                                                                           },200f,GameFont.Medium,TextAnchor.MiddleCenter))),
			                   200f,200f)
			                   ));
			
			pages.Add("viewOrder", new PageRenderer(f=>titleSelectOrder, null, ()=>"selectOrder", null)
	                    .AddChild(new VerticalSplitRenderer(new RowLayoutRenderer()
			                    .AddChild(new LabelRenderer(labelRecipe,-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    .AddChild(new LabelRenderer(f=>{
			                                                     	CraftOrder order = f.flowScope["selectItem"] as CraftOrder; 
			                                                     	if ( order != null ) {
			                                                     		return order.recipe == null ? "error" : order.recipe.label;
			                                                     	}
			                                                     	return "error";
			                                                     },-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    .AddChild(new LabelRenderer(labelQuality,-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    .AddChild(new HideRenderer(f=>{
			                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                               	return order == null || !order.HasQuality();
			                                               },
			                                               new LabelRenderer(f => (f.flowScope["selectItem"] as CraftOrder).quality.GetLabel(),-1f,GameFont.Medium,TextAnchor.MiddleCenter)))
			                    .AddChild(new LabelRenderer(labelStuff,-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    .AddChild(new LabelRenderer(f=>{
			                                                	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                	if ( order != null ) {
			                                                		if ( order.HasStuff() )
			                                                			return order.ingredients.First(p=>p.First.IsStuff).First.label;
			                                                		return labelNone;
			                                                	}
			                                                	return "error";
			                                                },-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    
			                    .AddChild(new LabelRenderer(labelAmount,-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    .AddChild(new LabelRenderer(f=>{
			                                                	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
	                                                          	if ( order != null ) {
	                                                          		return order.amount.ToString("0");
	                                                          	}
	                                                          	return "error";
	                                                          },-1f,GameFont.Medium,TextAnchor.MiddleCenter)),
			                    new RowLayoutRenderer()
			                    .AddChild(new LabelRenderer(labelIngredients,-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    .AddChild(new ListRenderer(f=>{
			                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                               	if ( order != null ) return order.ingredients;
			                                               	return new List<Pair<ThingDef, IngredientCount>>();
			                                               }, -1, QOLMod.LineHeight(GameFont.Medium)*3)
			                              .AddChild(new LabelRenderer(f => {
                                                                        	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                                               	Pair<ThingDef, IngredientCount> p = ((Pair<ThingDef, IngredientCount>)((IterationItem)f.pageScope["curItem"]).curItem);
			                                                               	ThingDef tDef = p.First;
			                                                               	int crf = p.Second.CountRequiredOfFor(p.First, order.recipe);
			                                                               	return tDef.label+" "+crf.ToString();
			                                                               },200f,GameFont.Medium,TextAnchor.MiddleCenter)))
			                    
			                    .AddChild(new LabelRenderer(labelProducts,-1f,GameFont.Medium,TextAnchor.MiddleCenter))
			                    .AddChild(new ListRenderer(f=>{
			                                               	CraftOrder order = f.flowScope["selectItem"] as CraftOrder;
			                                               	if ( order != null ) return order.products;
			                                               	return new List<ThingDefCountClass>();
			                                               }, -1, QOLMod.LineHeight(GameFont.Medium)*3)
			                                               .AddChild(new LabelRenderer(f=>{
			                                                                           	ThingDefCountClass tdcc = ((ThingDefCountClass)((IterationItem)f.pageScope["curItem"]).curItem);
			                                                                           	if ( tdcc != null ) return tdcc.thingDef.label+" "+tdcc.count;
			                                                                           	return "error";
			                                                                           },200f,GameFont.Medium,TextAnchor.MiddleCenter))),
			                    200f,200f)
			                   ));
			
			base.flow = new Flow(pages,"menu");
		}
		
		override public void PreOpen() {
			base.PreOpen();
		}
		
		public override void PostClose() {
			base.PostClose();
			UI.UnfocusCurrentControl();
			city.Refresh();
		}
	}
}
