﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 09.02.2019
 * Time: 02:23
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Linq;
using System.Collections.Generic;
using RIMMSqol.remnantcolony.actions;
using UnityEngine;
using Verse;
using RimWorld;

namespace RIMMSqol.remnantcolony.listeners
{
	public interface IQOLCityListener_ConversionConfiguration {
		QOLCityStateContribution InputPerCycle{get;set;}
		QOLCityStateContribution OutputPerCycle{get;set;}
		float LabourPerCycle{get;set;}
		int MaxCyclePerInstance{get;set;}
		bool OnlyNecessaryProduction{get;set;}
		int CyclesAllocated(QOLCity city, IQOLCityListenerOwner owner, QOLCityState state);
		int ActiveInstances(QOLCity city, IQOLCityListenerOwner owner, QOLCityState state);
		void CyclesPerformed(QOLCity city, IQOLCityListenerOwner owner, QOLCityState state, int cyclesPerformed);
	}
	
	public abstract class QOLCityListener_ConversionConfiguration_ActionBase : QOLCityActionConfiguration, IQOLCityListener_ConversionConfiguration {
		protected QOLCityStateContribution inputPerCycle;
		protected QOLCityStateContribution outputPerCycle;
		protected float labourPerCycle;
		protected int maxCyclePerInstance;
		protected bool onlyNecessaryProduction;
		// disable ConvertToAutoProperty
		public virtual QOLCityStateContribution InputPerCycle {get{return inputPerCycle;}set{inputPerCycle = value;}}
		public virtual QOLCityStateContribution OutputPerCycle {get{return outputPerCycle;}set{outputPerCycle = value;}}
		public virtual float LabourPerCycle {get{return labourPerCycle;}set{labourPerCycle = value;}}
		public virtual int MaxCyclePerInstance {get{return maxCyclePerInstance;}set{maxCyclePerInstance = value;}}
		public virtual bool OnlyNecessaryProduction {get{return onlyNecessaryProduction;}set{onlyNecessaryProduction = value;}}
		public virtual int ActiveInstances(QOLCity city, IQOLCityListenerOwner owner, QOLCityState state) { return 1; }
		public virtual int CyclesAllocated(QOLCity city, IQOLCityListenerOwner owner, QOLCityState state) { return 1; }
		public virtual void CyclesPerformed(QOLCity city, IQOLCityListenerOwner owner, QOLCityState state, int cyclesPerformed) {}
	}
	
	public class QOLCityListener_Conversion : QOLCityListener
	{
		protected bool initialized = false;
		protected bool hasInput, hasOutput;
		
		public QOLCityListener_Conversion(QOLCityListenerDef def,IQOLCityListenerOwner owner) : base(def,owner) {}
		
		public override void OnEvent(QOLCity city, QOLCityEvent e) {
			IQOLCityListener_ConversionConfiguration conf;
			if ( owner is QOLCityAction ) {
				conf = ((QOLCityAction)owner).def.action as IQOLCityListener_ConversionConfiguration;
			} else if ( owner is QOLCityProject ) {
				conf = ((QOLCityProject)owner).def as IQOLCityListener_ConversionConfiguration;
			} else {
				conf = owner as IQOLCityListener_ConversionConfiguration;
			}
			if ( conf != null ) {
				QOLCityState state;
				QOLCityEvent_State stateEvent = e as QOLCityEvent_State;
				if ( stateEvent != null ) {
					state = stateEvent.State;
				} else {
					state = city.stateLastSimulatedTick;
				}
				
				int cyclesAllocated = conf.CyclesAllocated(city,owner,state);
				if ( cyclesAllocated > 0 ) {
					if ( conf != null ) {
						if ( !initialized ) {
							//We can not consume labour while contributing in the same step.
							if ( conf.InputPerCycle != null && conf.InputPerCycle.needs.RemoveAll(nc => nc.def == city.defCity.labourDef) > 0 ) {
								Log.Warning("Configuration for listener "+def.defName+" is using labour as input! Entries will be ignored.");
							}
							
							hasInput = !conf.InputPerCycle.NullOrEmpty();
							hasOutput = !conf.OutputPerCycle.NullOrEmpty();
							
							initialized = true;
						}
						
						int cyclesEffectively = cyclesAllocated;
						int activeInstances = conf.ActiveInstances(city,owner,state);
						QOLCityAction action = owner as QOLCityAction;
						if ( action != null ) {
							activeInstances = city.projects.Count(p=>p.IsActive && p.def.actions.Contains(action.def));
						} else {
							activeInstances = 1;
						}
						cyclesEffectively = Math.Min(cyclesEffectively,activeInstances * conf.MaxCyclePerInstance);
						//Log.Message("cycles: "+cyclesEffectively);
						//How many cycles can be afforded with the labour excess?
						float availableLabour = city.LabourBalance(state);
						float labourNeededPerCycle = conf.LabourPerCycle;
						if ( labourNeededPerCycle > 0f ) cyclesEffectively = Math.Min(cyclesEffectively,Mathf.FloorToInt(availableLabour/labourNeededPerCycle));
						//Log.Message("cycles after labour: "+cyclesEffectively);
						
						//How often can the input state be applied without having a needed resource/need-balance negative?
						if ( hasInput ) {
							cyclesEffectively = state.Accommodate(conf.InputPerCycle, cyclesEffectively);
						}
						//Log.Message("cycles after accommodate: "+cyclesEffectively);
						
						//determine which needs and resources are in the output and then check those for positive values after each output added and break if all are positive.
						Func<bool> allOutputIsSatisfied;
						if ( conf.OnlyNecessaryProduction && hasOutput ) {
							allOutputIsSatisfied = compileAllOutputIsSatisfied(state,conf.OutputPerCycle);
						} else {
							allOutputIsSatisfied = () => false;
						}
						ProvidedAndRequired par = city.LabourPaR(stateEvent.State);
						int cyclesPerformed = 0;
						for ( int i = 0; i < cyclesEffectively; i++ ) {
							if ( hasInput ) state.Add(conf.InputPerCycle);
							if ( hasOutput ) state.Add(conf.OutputPerCycle);
							par.required.flat += conf.LabourPerCycle;
							cyclesPerformed++;
							//Log.Message("convert cycle");
							if ( allOutputIsSatisfied() ) {
								//Log.Message("convert outputs satisfied");
								break;
							}
						}
						conf.CyclesPerformed(city,owner,state,cyclesPerformed);
					}
				}
			} else {
				Log.Error("Listener \""+typeof(QOLCityListener_Conversion).Name+"\" was registered on an incompatible owner \""+owner+"\"!");
			}
		}
		
		protected Func<bool> compileAllOutputIsSatisfied(QOLCityState state, QOLCityStateContribution con) {
			List<NumericAdjustment> nas = new List<NumericAdjustment>();
			List<ProvidedAndRequired> pars = new List<ProvidedAndRequired>();
			foreach ( QOLCityStateContributionResource cr in con.resources ) {
				NumericAdjustment na;
				if ( state.resources.TryGetValue(cr.def, out na) ) {
					nas.Add(na);
				}
			}
			foreach ( QOLCityStateContributionThing ct in con.things ) {
				NumericAdjustment na;
                KeyValuePair<ThingDef, QualityCategory> kvp = new KeyValuePair<ThingDef, QualityCategory>(ct.def, ct.quality);
                if ( state.things.TryGetValue(kvp, out na) ) {
                    //test that we have more than enough things in the current cycle and the city.
                    int stockCount = state.city.GetDirectlyHeldThings().Where(t => ct.def != null && t.def.defName == ct.def.defName).Sum(t => t.stackCount);
                    na.flat += stockCount;
                    nas.Add(na);
                }
			}
			foreach ( QOLCityStateContributionNeed cn in con.needs ) {
				ProvidedAndRequired par;
				if ( state.needs.TryGetValue(cn.def, out par) ) {
					pars.Add(par);
				}
			}
			
			return ()=>nas.TrueForAll(na=>na.Value>0f) && pars.TrueForAll(par=>par.provided.Value>=par.required.Value);
		}
	}
}
