﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 28.10.2019
 * Time: 06:23
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using RimWorld;
using RimWorld.Planet;
using UnityEngine;
using Verse;

namespace RIMMSqol.storyteller
{
	/// <summary>
	/// Description of StorytellerTree.
	/// </summary>
	public class StorytellerTree : IExposable
	{
		protected Dictionary<Type,StorytellerTreeComp> comps;
		protected StorytellerTreeNode root;
		protected string labelTurn, labelInterval, labelDay;
		public StorytellerTreeQueue incidentQueue = new StorytellerTreeQueue();
		public int lastInterval, lastDay;
		public List<IIncidentTarget> lastTargets;
		
		public StorytellerTreeNode Root {
			get { return root; }
		}
		
		public StorytellerTree() {
			root = new StorytellerTreeNode_Root();
			InitializeStorytellerComps();
		}
		
		public void initializeTree(StorytellerTreeDef td) {
			root = td.spawnTree();
		}
		
		public virtual void ExposeData() {
			Scribe_Deep.Look<StorytellerTreeNode>(ref root, "root", new object[0]);
			Scribe_Deep.Look<StorytellerTreeQueue>(ref incidentQueue, "incidentQueue", new object[0]);
			Scribe_Values.Look<int>(ref lastInterval, "lastInterval");
			
			if (Scribe.mode == LoadSaveMode.ResolvingCrossRefs) {
				InitializeStorytellerComps();
			}
		}
		
		protected void InitializeStorytellerComps() {
			comps = new Dictionary<Type, StorytellerTreeComp>();
			typeof(StorytellerTreeComp).AllSubclassesNonAbstract();
			foreach ( Type comp in typeof(StorytellerTreeComp).AllSubclassesNonAbstract() ) {
				StorytellerTreeComp compInstance = (StorytellerTreeComp)Activator.CreateInstance(comp);
				compInstance.AttachTree(this);
				comps.Add(comp,compInstance);
			}
		}
		
		public T getComp<T>() where T : StorytellerTreeComp {
			StorytellerTreeComp res;
			if ( comps.TryGetValue(typeof(T),out res) ) {
				return (T)res;
			}
			return null;
		}
		
		protected List<IIncidentTarget> CompileAllIncidentTargets() {
			List<IIncidentTarget> targets = new List<IIncidentTarget>();
			foreach ( Map map in Find.Maps ) {
				targets.Add(map);
			}
			foreach ( Caravan caravan in Find.WorldObjects.Caravans ) {
				if ( caravan.IsPlayerControlled ) targets.Add(caravan);
			}
			targets.Add(Find.World);
			return targets;
		}
		
		public virtual void TryStartIncident() {
			incidentQueue.IncidentQueueTick();
			
			int currentInterval = Find.TickManager.TicksGame / 1000; 
			if ( currentInterval > lastInterval ) {
				lastDay = GenDate.DaysPassed;
				lastInterval = currentInterval;
				lastTargets = CompileAllIncidentTargets();
				
				foreach ( StorytellerTreeComp comp in comps.Values ) {
					comp.BeginInterval();
				}
				if ( root != null ) {
					foreach ( StorytellerTreeIncident fi in root.TryStartIncident(this) ) {
						TryFire(fi);
					}
				}
				foreach ( StorytellerTreeComp comp in comps.Values ) {
					comp.EndInterval();
				}
				
				lastTargets.Clear();
			}
		}
		
		protected bool TryFire(StorytellerTreeIncident fi) {
			if ( fi.delay > 0 ) {
				fi.NotifyFiredOrQueued(this);
				incidentQueue.Add(new QueuedIncident(fi.incident, Find.TickManager.TicksGame + fi.delay, fi.retryDurationTicks));
				//Log.Message("Queued "+fi.incident.def.label);
				fi.Dispose();
				return true;
			} else {
				if ((fi.incident.parms.forced || fi.incident.def.Worker.CanFireNow(fi.incident.parms)) && fi.incident.def.Worker.TryExecute(fi.incident.parms)) {
					fi.NotifyFiredOrQueued(this);
					fi.incident.parms.target.StoryState.Notify_IncidentFired(fi.incident);
					//Log.Message("Fired "+fi.incident.def.label);
					fi.Dispose();
					return true;
				}
				return false;
			}
		}
		
		public void Dispose() {
			if ( root != null ) root.Dispose();
			if ( lastTargets != null ) lastTargets.Clear();
			if ( incidentQueue != null ) incidentQueue.Clear();
			labelTurn = labelInterval = labelDay = null;
		}
		
		public virtual void DoSettingsContent(Rect rect,  out float requiredWidth, out float requiredHeight) {
			if ( labelTurn == null ) {
				TaggedString ts; "Turn".TryTranslate(out ts); labelTurn = ts;
			}
			if ( labelInterval == null ) {
				TaggedString ts; "Interval".TryTranslate(out ts); labelInterval = ts;
			}
			if ( labelDay == null ) {
				TaggedString ts; "Day".TryTranslate(out ts); labelDay = ts;
			}
			int turn = Find.TickManager.TicksGame;
			int interval = Find.TickManager.TicksGame / 1000;
			int day = GenDate.DaysPassed;
			Rect r = new Rect(rect.x, rect.y, 200f, Text.LineHeight);
			Widgets.Label(r,labelTurn);r.x += 208f; Widgets.Label(r,turn.ToString("D"));
			r.y += Text.LineHeight + 4f;r.x = rect.x;
			Widgets.Label(r,labelInterval);r.x += 208f; Widgets.Label(r,interval.ToString("D"));
			r.y += Text.LineHeight + 4f;r.x = rect.x;
			Widgets.Label(r,labelDay);r.x += 208f; Widgets.Label(r,day.ToString("D"));
			requiredWidth = 416f;
			requiredHeight = r.height + r.y - rect.y;
		}
	}
}
