﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 28.10.2019
 * Time: 06:23
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Reflection;
using RIMMSqol.genericSettings;
using RimWorld;
using UnityEngine;
using Verse;

namespace RIMMSqol.storyteller
{
	/// <summary>
	/// Description of StorytellerTreeNode.
	/// </summary>
	public abstract class StorytellerTreeNode : IExposable
	{
		protected StorytellerTreeNode parent;
		protected List<StorytellerTreeNode> children;
		public string label;
		
		public virtual bool IsLeaf { get { return false; } }
		
		public StorytellerTreeNode Parent {
			get { return parent; }
		}
		
		public StorytellerTreeNode FirstSibling {
			get { return children != null ? children.FirstOrDefault() : null; }
		}
		
		public StorytellerTreeNode NextSibling(StorytellerTreeNode sibling) {
			int indx = children.IndexOf(sibling) + 1;
			if ( indx == 0 || indx >= children.Count ) return null;
			return children[indx];
		}
		
		protected virtual void InitSettings(SettingsProperties p) {
			p.fields.Add(new SettingsFieldPropertiesPrimitiveString<StorytellerTreeNode>("Label","Label",1,null,t=>t.label));
			p.visualization = new SettingsVisualizationEdit(f=>null);
		}
		
		protected virtual void AdoptSettings(SettingsInstance si, StorytellerTreeNode obj) {
			obj.label = si.get<string>("Label");
		}
		
		protected virtual SettingsProperties GetSettings() {
			string propertiesId = this.GetType().Name;
			SettingsProperties p = SettingsStorage.getSettingsProperties(propertiesId);
			if ( p == null ) {
				p = SettingsStorage.getOrInitSettingsProperties(propertiesId);
				InitSettings(p);
			}
			return p;
		}
		
		protected StorytellerTreeNode() : this(null) {}
		
		protected StorytellerTreeNode(StorytellerTreeNode parent) {
			this.parent = parent;
			children = null;
			
			if ( this.parent != null ) {
				this.parent.AddChild(this);
			}
		}
		
		protected virtual void AddChild(StorytellerTreeNode child) {
			if ( children == null ) children = new List<StorytellerTreeNode>();
			children.Add(child);
		}
		
		public virtual void RemoveChild(StorytellerTreeNode child) {
			if ( children == null || child == null ) return;
			if ( children.Remove(child) ) {
				child.parent = null;
			}
		}
		
		public virtual void MoveChildForward(StorytellerTreeNode child) {
			if ( children == null || child == null ) return;
			int indx = children.IndexOf(child);
			if ( indx > 0 ) {
				StorytellerTreeNode n = children[indx-1];
				children[indx-1] = child;
				children[indx] = n;
			}
		}
		
		public virtual void MoveChildBackwards(StorytellerTreeNode child) {
			if ( children == null || child == null ) return;
			int indx = children.IndexOf(child);
			if ( indx < children.Count-1 ) {
				StorytellerTreeNode n = children[indx+1];
				children[indx+1] = child;
				children[indx] = n;
			}
		}
		
		public virtual IEnumerable<StorytellerTreeIncident> TryStartIncident(StorytellerTree tree) {
			if ( !children.NullOrEmpty() ) {
				foreach ( StorytellerTreeNode child in children ) {
					foreach ( StorytellerTreeIncident fi in child.TryStartIncident(tree) ) {
						yield return fi;
					}
				}
			}
			yield break;
		}
		
		public virtual void DoSettingsDialog(StorytellerTree tree) {
			
		}
		
		public virtual void Dispose() {
			if ( !children.NullOrEmpty() ) {
				foreach ( StorytellerTreeNode n in children ) {
					n.Dispose();
				}
				children.Clear();
			}
			label = null;
			children = null;
		}
		
		public virtual void ExposeData() {
			Scribe_Values.Look<string>(ref label, "label");
			Scribe_Collections.Look<StorytellerTreeNode>(ref children, "children", LookMode.Deep);
			if (Scribe.mode == LoadSaveMode.ResolvingCrossRefs) {
				if ( !children.NullOrEmpty() ) {
					foreach ( StorytellerTreeNode child in children ) {
						child.parent = this;
					}
				}
			}
		}
		
		public bool IsDescendantOf(StorytellerTreeNode possibleAncestor) {
			StorytellerTreeNode n = this;
			while ( n != null ) {
				if ( n == possibleAncestor ) return true;
				n = n.parent;
			}
			return false;
		}
		
		public bool IsAncestorOf(StorytellerTreeNode possibleDescendant) {
			StorytellerTreeNode n = possibleDescendant;
			while ( n != null ) {
				if ( n == this ) return true;
				n = n.parent;
			}
			return false;
		}
		
		public void DeepCopySubtree(StorytellerTreeNode parent) {
			if ( IsAncestorOf(parent) ) return;
			
			DeepCopySubtree(this,parent);
		}
		
		protected void DeepCopySubtree(StorytellerTreeNode toCopy, StorytellerTreeNode parent) {
			StorytellerTreeNode newParent = (StorytellerTreeNode)Activator.CreateInstance(toCopy.GetType(),parent);
			
			toCopy.AdoptSettings(new SettingsInstance(toCopy.GetSettings().id,toCopy),newParent);
			
			if ( toCopy.children != null ) {
				foreach (StorytellerTreeNode child in toCopy.children) {
					DeepCopySubtree(child,newParent);
				}
			}
		}
		
		protected static void loadCurveFromPoints(List<float> curvePoints, out SimpleCurve curve) {
			if ( !curvePoints.NullOrEmpty() ) {
				curve = new SimpleCurve();
				int limit = curvePoints.Count-1;
				for ( int i = 0; i < limit; i+=2 ) {
					curve.Add(curvePoints[i],curvePoints[i+1]);
				}
			} else curve = null;
		}
		
		protected static List<float> curvePointsAsList(SimpleCurve curve) {
			List<float> curvePoints = new List<float>();
			if ( curve != null ) {
				foreach ( CurvePoint p in curve.Points ) {
					curvePoints.Add(p.x);
					curvePoints.Add(p.y);
				}
			}
			return curvePoints;
		}
		
		protected IncidentParms generateParameter(IncidentCategoryDef category, IIncidentTarget target) {
			IncidentParms p = StorytellerUtility.DefaultParmsNow(category, target);
			if ( p.questScriptDef == null && category.needsParmsPoints ) {
				p.questScriptDef = NaturalRandomQuestChooser.ChooseNaturalRandomQuest(p.points, target);
			}
			return p;
		}
	}
	
	public abstract class StorytellerTreeNodeConfiguration {
		public string label;
		public List<StorytellerTreeNodeConfiguration> children;
		public abstract StorytellerTreeNode CreateNode(StorytellerTreeNode parent);
	}
}
