﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 16.11.2019
 * Time: 19:04
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using RIMMSqol.genericSettings;
using RimWorld;
using Verse;

namespace RIMMSqol.storyteller
{
	/// <summary>
	/// Description of StorytellerTreeNode_Loop.
	/// </summary>
	public class StorytellerTreeNode_Loop : StorytellerTreeNode 
	{
		public ThingRequestGroup thingRequestGroup = ThingRequestGroup.Undefined;
		
		public int repetitionsMin = 1, repetitionsMax = 1;
		
		public StorytellerTreeNode_Loop() : this(null) { }
		
		public StorytellerTreeNode_Loop(StorytellerTreeNode parent) : base(parent) { }
		
		protected void AddToLoop(ref Dictionary<IIncidentTarget,float> oldLoopCounters, StorytellerTreeComp_TargetFilter targetFilter, IIncidentTarget target, float loop) {
			if ( oldLoopCounters == null ) oldLoopCounters = new Dictionary<IIncidentTarget, float>();
			if ( !oldLoopCounters.ContainsKey(target) ) {
				oldLoopCounters.Add(target,targetFilter.getLoopCounter(target));
			}
			targetFilter.AddLoopMultiplier(target,loop);
		}
		
		public override IEnumerable<StorytellerTreeIncident> TryStartIncident(StorytellerTree tree) {
			Dictionary<IIncidentTarget,float> oldLoopCounters = null;
			
			if ( thingRequestGroup != ThingRequestGroup.Undefined ) {
				StorytellerTreeComp_TargetFilter targetFilter = tree.getComp<StorytellerTreeComp_TargetFilter>();
				foreach ( IIncidentTarget target in targetFilter.FilteredTargetsNoLoop ) {
					Map map = target as Map;
					if ( map != null ) {
						int count = map.listerThings.ThingsInGroup(thingRequestGroup).Count;
						AddToLoop(ref oldLoopCounters, targetFilter, map, count);
					}
				}
			}
			
			if ( repetitionsMin > 1 || repetitionsMax > 1 ) {
				StorytellerTreeComp_TargetFilter targetFilter = tree.getComp<StorytellerTreeComp_TargetFilter>();
				foreach ( IIncidentTarget target in targetFilter.FilteredTargetsNoLoop ) {
					int repetitions = Rand.RangeInclusive(repetitionsMin, repetitionsMax);
					AddToLoop(ref oldLoopCounters, targetFilter, target, repetitions);
				}
			}
			
			foreach ( StorytellerTreeIncident fi in base.TryStartIncident(tree) ) yield return fi;
			
			if ( oldLoopCounters != null ) {
				StorytellerTreeComp_TargetFilter targetFilter = tree.getComp<StorytellerTreeComp_TargetFilter>();
				foreach ( KeyValuePair<IIncidentTarget,float> p in oldLoopCounters ) {
					targetFilter.setLoopCounter(p.Key,p.Value);
				}
			}
		}
		
		public override void ExposeData() {
			base.ExposeData();
			Scribe_Values.Look<ThingRequestGroup>(ref thingRequestGroup, "thingRequestGroup", ThingRequestGroup.Undefined);
			Scribe_Values.Look<int>(ref repetitionsMin, "repetitionsMin", 1);
			Scribe_Values.Look<int>(ref repetitionsMax, "repetitionsMax", 1);
			
			if (Scribe.mode == LoadSaveMode.ResolvingCrossRefs) {
				repetitionsMin = Math.Max(1,repetitionsMin);
				repetitionsMax = Math.Max(repetitionsMin,repetitionsMax);
			}
		}
		
		protected override void InitSettings(SettingsProperties p) {
			base.InitSettings(p);
			p.fields.Add(new SettingsFieldPropertiesSelectable("ThingRequestGroup","Thing Count",10,typeof(string),Enum.GetName(typeof(ThingRequestGroup),ThingRequestGroup.Undefined),
			                                                   t=>Enum.GetName(typeof(ThingRequestGroup),((StorytellerTreeNode_Loop)t).thingRequestGroup),
                                                       f=>from ThingRequestGroup d in Enum.GetValues(typeof(ThingRequestGroup)) orderby Enum.GetName(typeof(ThingRequestGroup),d) select (object)Enum.GetName(typeof(ThingRequestGroup),d),
			                                             f=>def=>{ ((ISettingsInstance)f.flowScope["edit"]).set<string>("ThingRequestGroup",(string)def); return true; },
			                                             o=>(string)o,false));
			p.fields.Add(new SettingsFieldPropertiesPrimitiveInt<StorytellerTreeNode_Loop>("RepetitionsMin","Repetitions Min",20,1,t=>t.repetitionsMin,1));
			p.fields.Add(new SettingsFieldPropertiesPrimitiveInt<StorytellerTreeNode_Loop>("RepetitionsMax","Repetitions Max",30,1,t=>t.repetitionsMax,1));
		}
		
		protected override void AdoptSettings(SettingsInstance si, StorytellerTreeNode obj) {
			base.AdoptSettings(si, obj);
			StorytellerTreeNode_Loop sttn = (StorytellerTreeNode_Loop)obj;
			sttn.thingRequestGroup = (ThingRequestGroup)Enum.Parse(typeof(ThingRequestGroup),si.get<string>("ThingRequestGroup"));
           	sttn.repetitionsMin = si.get<int>("RepetitionsMin");
           	sttn.repetitionsMax = si.get<int>("RepetitionsMax");
           	
           	sttn.repetitionsMin = Math.Max(1,sttn.repetitionsMin);
			sttn.repetitionsMax = Math.Max(sttn.repetitionsMin,sttn.repetitionsMax);
		}
		
		public override void DoSettingsDialog(StorytellerTree tree) {
			Find.WindowStack.Add(new Dialog_EditObject<StorytellerTreeNode_Loop>(this,GetSettings().id,AdoptSettings));
		}
	}
	
	public class StorytellerTreeNodeConfiguration_Loop : StorytellerTreeNodeConfiguration {
		public ThingRequestGroup thingRequestGroup = ThingRequestGroup.Undefined;
		public int repetitionsMin = 1, repetitionsMax = 1;
		
		public override StorytellerTreeNode CreateNode(StorytellerTreeNode parent) {
			StorytellerTreeNode_Loop n = new StorytellerTreeNode_Loop(parent);
			n.label = label;
			n.thingRequestGroup = thingRequestGroup;
			n.repetitionsMin = repetitionsMin;
			n.repetitionsMax = repetitionsMax;
			return n;
		}
	}
}
